# -*- coding: utf-8 -*-
from __future__ import annotations

import asyncio
import json
from typing import Any, Dict, List, Optional, Tuple
from uuid import UUID

from langchain.callbacks import AsyncIteratorCallbackHandler
from langchain.schema import AgentAction, AgentFinish
from langchain_core.outputs import LLMResult

from langchain_glm.agent_toolkits import BaseToolOutput
from langchain_glm.utils import History


def dumps(obj: Dict) -> str:
    return json.dumps(obj, ensure_ascii=False)


class AgentStatus:
    chain_start: int = 0
    llm_start: int = 1
    llm_new_token: int = 2
    llm_end: int = 3
    agent_action: int = 4
    agent_finish: int = 5
    tool_start: int = 6
    tool_end: int = 7
    error: int = -1
    chain_end: int = -999


class AgentExecutorAsyncIteratorCallbackHandler(AsyncIteratorCallbackHandler):
    def __init__(self):
        super().__init__()
        self.queue = asyncio.Queue()
        self.done = asyncio.Event()
        self.out = False
        self.intermediate_steps: List[Tuple[AgentAction, BaseToolOutput]] = []
        self.outputs: Dict[str, Any] = {}

    async def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        data = {
            "status": AgentStatus.llm_start,
            "text": "",
        }
        self.out = False
        self.done.clear()
        self.queue.put_nowait(dumps(data))

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        special_tokens = ["\nAction:", "\nObservation:", "<|observation|>"]
        for stoken in special_tokens:
            if stoken in token:
                before_action = token.split(stoken)[0]
                data = {
                    "status": AgentStatus.llm_new_token,
                    "text": before_action + "\n",
                }
                self.queue.put_nowait(dumps(data))
                self.out = False
                break

        if token is not None and token != "" and not self.out:
            data = {
                "run_id": str(kwargs["run_id"]),
                "status": AgentStatus.llm_new_token,
                "text": token,
            }
            self.queue.put_nowait(dumps(data))

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.llm_start,
            "text": "",
        }
        self.done.clear()
        self.queue.put_nowait(dumps(data))

    async def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        data = {
            "run_id": str(kwargs["run_id"]),
            "status": AgentStatus.llm_end,
            "text": response.generations[0][0].message.content,
        }

        self.queue.put_nowait(dumps(data))

    async def on_llm_error(
        self, error: Exception | KeyboardInterrupt, **kwargs: Any
    ) -> None:
        data = {
            "status": AgentStatus.error,
            "text": str(error),
        }
        self.queue.put_nowait(dumps(data))

    async def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.tool_start,
            "tool": serialized["name"],
            "tool_input": input_str,
        }
        self.done.clear()
        self.queue.put_nowait(dumps(data))

    async def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Run when tool ends running."""
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.tool_end,
            "tool": kwargs["name"],
            "tool_output": output,
        }
        self.queue.put_nowait(dumps(data))

    async def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Run when tool errors."""
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.error,
            "tool_output": str(error),
            "is_error": True,
        }

        self.queue.put_nowait(dumps(data))

    async def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.agent_action,
            "action": {
                "tool": action.tool,
                "tool_input": action.tool_input,
                "log": action.log,
            },
        }
        self.queue.put_nowait(dumps(data))

    async def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        if "Thought:" in finish.return_values["output"]:
            finish.return_values["output"] = finish.return_values["output"].replace(
                "Thought:", ""
            )

        data = {
            "run_id": str(run_id),
            "status": AgentStatus.agent_finish,
            "finish": {
                "return_values": finish.return_values,
                "log": finish.log,
            },
        }

        self.queue.put_nowait(dumps(data))

    async def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        if "agent_scratchpad" in inputs:
            del inputs["agent_scratchpad"]
        if "chat_history" in inputs:
            inputs["chat_history"] = [
                History.from_message(message).to_msg_tuple()
                for message in inputs["chat_history"]
            ]
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.chain_start,
            "inputs": inputs,
            "parent_run_id": parent_run_id,
            "tags": tags,
            "metadata": metadata,
        }

        self.done.clear()
        self.out = False
        self.queue.put_nowait(dumps(data))

    async def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain errors."""
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.error,
            "error": str(error),
        }
        self.queue.put_nowait(dumps(data))

    async def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        if "intermediate_steps" in outputs:
            self.intermediate_steps = outputs["intermediate_steps"]
            self.outputs = outputs
            del outputs["intermediate_steps"]
        data = {
            "run_id": str(run_id),
            "status": AgentStatus.chain_end,
            "outputs": outputs,
            "parent_run_id": parent_run_id,
            "tags": tags,
        }
        self.queue.put_nowait(dumps(data))
        self.out = True
        # self.done.set()



# -*- coding: utf-8 -*-
from __future__ import annotations

import asyncio
import json
import logging
import time
from abc import abstractmethod
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
)

from langchain.agents.agent import Agent, AgentExecutor, AgentOutputParser
from langchain.agents.output_parsers.tools import ToolAgentAction
from langchain.agents.tools import InvalidTool
from langchain.utilities.asyncio import asyncio_timeout
from langchain_core.agents import AgentAction, AgentFinish, AgentStep
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    AsyncCallbackManagerForToolRun,
    BaseCallbackManager,
    CallbackManagerForChainRun,
    CallbackManagerForToolRun,
    Callbacks,
)
from langchain_core.tools import BaseTool
from langchain_core.utils import get_color_mapping

from langchain_glm.agent_toolkits import AdapterAllTool
from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agents.output_parsers.code_interpreter import (
    CodeInterpreterAgentAction,
)
from langchain_glm.agents.output_parsers.drawing_tool import DrawingToolAgentAction
from langchain_glm.agents.output_parsers.web_browser import WebBrowserAgentAction

logger = logging.getLogger(__name__)


class ZhipuAiAllToolsAgentExecutor(AgentExecutor):
    def _call(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """Run text through and get agent response."""
        # Construct a mapping of tool name to tool for easy lookup
        name_to_tool_map = {tool.name: tool for tool in self.tools}
        # We construct a mapping from each tool to a color, used for logging.
        color_mapping = get_color_mapping(
            [tool.name for tool in self.tools], excluded_colors=["green", "red"]
        )
        intermediate_steps: List[Tuple[AgentAction, str]] = []
        # Let's start tracking the number of iterations and time elapsed
        iterations = 0
        time_elapsed = 0.0
        start_time = time.time()
        # We now enter the agent loop (until it returns something).
        while self._should_continue(iterations, time_elapsed):
            next_step_output = self._take_next_step(
                name_to_tool_map,
                color_mapping,
                inputs,
                intermediate_steps,
                run_manager=run_manager,
            )
            if isinstance(next_step_output, AgentFinish):
                return self._return(
                    next_step_output, intermediate_steps, run_manager=run_manager
                )

            intermediate_steps.extend(next_step_output)
            if len(next_step_output) == 1:
                next_step_action = next_step_output[0]
                # See if tool should return directly
                tool_return = self._get_tool_return(next_step_action)
                if tool_return is not None:
                    return self._return(
                        tool_return, intermediate_steps, run_manager=run_manager
                    )
            iterations += 1
            time_elapsed = time.time() - start_time
        output = self.agent.return_stopped_response(
            self.early_stopping_method, intermediate_steps, **inputs
        )
        return self._return(output, intermediate_steps, run_manager=run_manager)

    async def _acall(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, str]:
        """Run text through and get agent response."""
        # Construct a mapping of tool name to tool for easy lookup
        name_to_tool_map = {tool.name: tool for tool in self.tools}
        # We construct a mapping from each tool to a color, used for logging.
        color_mapping = get_color_mapping(
            [tool.name for tool in self.tools], excluded_colors=["green"]
        )
        intermediate_steps: List[Tuple[AgentAction, str]] = []
        # Let's start tracking the number of iterations and time elapsed
        iterations = 0
        time_elapsed = 0.0
        start_time = time.time()
        # We now enter the agent loop (until it returns something).
        try:
            async with asyncio_timeout(self.max_execution_time):
                while self._should_continue(iterations, time_elapsed):
                    next_step_output = await self._atake_next_step(
                        name_to_tool_map,
                        color_mapping,
                        inputs,
                        intermediate_steps,
                        run_manager=run_manager,
                    )
                    if isinstance(next_step_output, AgentFinish):
                        return await self._areturn(
                            next_step_output,
                            intermediate_steps,
                            run_manager=run_manager,
                        )

                    intermediate_steps.extend(next_step_output)
                    if len(next_step_output) >= 1:
                        # TODO: platform adapter status control, but langchain not output message info,
                        #   so where after paser instance object to let's DrawingToolAgentAction WebBrowserAgentAction
                        #   always output AgentFinish instance
                        continue_action = False
                        list1 = list(name_to_tool_map.keys())
                        list2 = [
                            AdapterAllToolStructType.WEB_BROWSER,
                            AdapterAllToolStructType.DRAWING_TOOL,
                        ]

                        exist_tools = list(set(list1) - set(list2))
                        for next_step_action, observation in next_step_output:
                            if next_step_action.tool in exist_tools:
                                continue_action = True
                                break

                        if not continue_action:
                            for next_step_action, observation in next_step_output:
                                if isinstance(next_step_action, DrawingToolAgentAction):
                                    tool_return = AgentFinish(
                                        return_values={"output": str(observation)},
                                        log=str(observation),
                                    )
                                    return await self._areturn(
                                        tool_return,
                                        intermediate_steps,
                                        run_manager=run_manager,
                                    )
                                elif isinstance(
                                    next_step_action, WebBrowserAgentAction
                                ):
                                    tool_return = AgentFinish(
                                        return_values={"output": str(observation)},
                                        log=str(observation),
                                    )
                                    return await self._areturn(
                                        tool_return,
                                        intermediate_steps,
                                        run_manager=run_manager,
                                    )

                    if len(next_step_output) == 1:
                        next_step_action = next_step_output[0]
                        # See if tool should return directly
                        tool_return = self._get_tool_return(next_step_action)
                        if tool_return is not None:
                            return await self._areturn(
                                tool_return, intermediate_steps, run_manager=run_manager
                            )

                    iterations += 1
                    time_elapsed = time.time() - start_time
                output = self.agent.return_stopped_response(
                    self.early_stopping_method, intermediate_steps, **inputs
                )
                return await self._areturn(
                    output, intermediate_steps, run_manager=run_manager
                )
        except (TimeoutError, asyncio.TimeoutError):
            # stop early when interrupted by the async timeout
            output = self.agent.return_stopped_response(
                self.early_stopping_method, intermediate_steps, **inputs
            )
            return await self._areturn(
                output, intermediate_steps, run_manager=run_manager
            )

    def _perform_agent_action(
        self,
        name_to_tool_map: Dict[str, BaseTool],
        color_mapping: Dict[str, str],
        agent_action: AgentAction,
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> AgentStep:
        if run_manager:
            run_manager.on_agent_action(agent_action, color="green")
        # Otherwise we lookup the tool
        if agent_action.tool in name_to_tool_map:
            tool = name_to_tool_map[agent_action.tool]
            return_direct = tool.return_direct
            color = color_mapping[agent_action.tool]
            tool_run_kwargs = self.agent.tool_run_logging_kwargs()
            if return_direct:
                tool_run_kwargs["llm_prefix"] = ""
            # We then call the tool on the tool input to get an observation
            # TODO: platform adapter tool for all  tools,
            #       view tools binding langchain_glm/agents/zhipuai_all_tools/base.py:188
            if agent_action.tool in AdapterAllToolStructType.__members__.values():
                observation = tool.run(
                    {
                        "agent_action": agent_action,
                    },
                    verbose=self.verbose,
                    color="red",
                    callbacks=run_manager.get_child() if run_manager else None,
                    **tool_run_kwargs,
                )
            else:
                observation = tool.run(
                    agent_action.tool_input,
                    verbose=self.verbose,
                    color=color,
                    callbacks=run_manager.get_child() if run_manager else None,
                    **tool_run_kwargs,
                )
        else:
            tool_run_kwargs = self.agent.tool_run_logging_kwargs()
            observation = InvalidTool().run(
                {
                    "requested_tool_name": agent_action.tool,
                    "available_tool_names": list(name_to_tool_map.keys()),
                },
                verbose=self.verbose,
                color=None,
                callbacks=run_manager.get_child() if run_manager else None,
                **tool_run_kwargs,
            )
        return AgentStep(action=agent_action, observation=observation)

    async def _aperform_agent_action(
        self,
        name_to_tool_map: Dict[str, BaseTool],
        color_mapping: Dict[str, str],
        agent_action: AgentAction,
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> AgentStep:
        if run_manager:
            await run_manager.on_agent_action(
                agent_action, verbose=self.verbose, color="green"
            )
        # Otherwise we lookup the tool
        if agent_action.tool in name_to_tool_map:
            tool = name_to_tool_map[agent_action.tool]
            return_direct = tool.return_direct
            color = color_mapping[agent_action.tool]
            tool_run_kwargs = self.agent.tool_run_logging_kwargs()
            if return_direct:
                tool_run_kwargs["llm_prefix"] = ""
            # We then call the tool on the tool input to get an observation
            # TODO: platform adapter tool for all  tools,
            #       view tools binding
            #       langchain_glm.agents.zhipuai_all_tools.base.ZhipuAIAllToolsRunnable.paser_all_tools
            if agent_action.tool in AdapterAllToolStructType.__members__.values():
                observation = await tool.arun(
                    {
                        "agent_action": agent_action,
                    },
                    verbose=self.verbose,
                    color="red",
                    callbacks=run_manager.get_child() if run_manager else None,
                    **tool_run_kwargs,
                )
            else:
                observation = await tool.arun(
                    agent_action.tool_input,
                    verbose=self.verbose,
                    color=color,
                    callbacks=run_manager.get_child() if run_manager else None,
                    **tool_run_kwargs,
                )
        else:
            tool_run_kwargs = self.agent.tool_run_logging_kwargs()
            observation = await InvalidTool().arun(
                {
                    "requested_tool_name": agent_action.tool,
                    "available_tool_names": list(name_to_tool_map.keys()),
                },
                verbose=self.verbose,
                color=None,
                callbacks=run_manager.get_child() if run_manager else None,
                **tool_run_kwargs,
            )
        return AgentStep(action=agent_action, observation=observation)



# -*- coding: utf-8 -*-
import json
import logging
from collections import deque
from json import JSONDecodeError
from typing import Any, Deque, Dict, List, Union

from langchain_core.agents import AgentAction, AgentActionMessageLog, AgentFinish
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ToolCall,
    ToolCallChunk,
)
from langchain_core.utils.json import (
    parse_partial_json,
)
from zhipuai.core import BaseModel

from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agents.output_parsers.base import (
    AllToolsMessageToolCall,
    AllToolsMessageToolCallChunk,
)
from langchain_glm.agents.output_parsers.code_interpreter import (
    _best_effort_parse_code_interpreter_tool_calls,
    _paser_code_interpreter_chunk_input,
)
from langchain_glm.agents.output_parsers.drawing_tool import (
    _best_effort_parse_drawing_tool_tool_calls,
    _paser_drawing_tool_chunk_input,
)
from langchain_glm.agents.output_parsers.function import (
    _best_effort_parse_function_tool_calls,
    _paser_function_chunk_input,
)
from langchain_glm.agents.output_parsers.web_browser import (
    _best_effort_parse_web_browser_tool_calls,
    _paser_web_browser_chunk_input,
)
from langchain_glm.chat_models.all_tools_message import ALLToolsMessageChunk

logger = logging.getLogger(__name__)


def paser_ai_message_to_tool_calls(
    message: BaseMessage,
):
    tool_calls = []
    if message.tool_calls:
        tool_calls = message.tool_calls
    else:
        if not message.additional_kwargs.get("tool_calls"):
            return AgentFinish(
                return_values={"output": message.content}, log=str(message.content)
            )
        # Best-effort parsing allready parsed tool calls
        for tool_call in message.additional_kwargs["tool_calls"]:
            if "function" == tool_call["type"]:
                function = tool_call["function"]
                function_name = function["name"]
                try:
                    args = json.loads(function["arguments"] or "{}")
                    tool_calls.append(
                        ToolCall(
                            name=function_name,
                            args=args,
                            id=tool_call["id"] if tool_call["id"] else "abc",
                        )
                    )
                except JSONDecodeError:
                    raise OutputParserException(
                        f"Could not parse tool input: {function} because "
                        f"the `arguments` is not valid JSON."
                    )
            elif tool_call["type"] in AdapterAllToolStructType.__members__.values():
                adapter_tool = tool_call[tool_call["type"]]

                tool_calls.append(
                    ToolCall(
                        name=tool_call["type"],
                        args=adapter_tool if adapter_tool else {},
                        id=tool_call["id"] if tool_call["id"] else "abc",
                    )
                )

    return tool_calls


def parse_ai_message_to_tool_action(
    message: BaseMessage,
) -> Union[List[AgentAction], AgentFinish]:
    """Parse an AI message potentially containing tool_calls."""
    if not isinstance(message, AIMessage):
        raise TypeError(f"Expected an AI message got {type(message)}")

    # TODO: parse platform tools built-in @langchain_glm.agents.zhipuai_all_tools.base._get_assistants_tool
    #   type in the future "function" or "code_interpreter"
    #   for @ToolAgentAction from langchain.agents.output_parsers.tools
    #   import with langchain.agents.format_scratchpad.tools.format_to_tool_messages
    actions: List = []
    tool_calls = paser_ai_message_to_tool_calls(message)
    if isinstance(tool_calls, AgentFinish):
        return tool_calls
    code_interpreter_action_result_stack: deque = deque()
    web_browser_action_result_stack: deque = deque()
    drawing_tool_result_stack: deque = deque()
    function_tool_result_stack: deque = deque()
    code_interpreter_chunk: List[
        Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]
    ] = []
    if message.tool_calls:
        if isinstance(message, ALLToolsMessageChunk):
            code_interpreter_chunk = _best_effort_parse_code_interpreter_tool_calls(
                message.tool_call_chunks
            )
    else:
        code_interpreter_chunk = _best_effort_parse_code_interpreter_tool_calls(
            tool_calls
        )

    if code_interpreter_chunk and len(code_interpreter_chunk) > 1:
        code_interpreter_action_result_stack = _paser_code_interpreter_chunk_input(
            message, code_interpreter_chunk
        )

    drawing_tool_chunk: List[
        Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]
    ] = []
    if message.tool_calls:
        if isinstance(message, ALLToolsMessageChunk):
            drawing_tool_chunk = _best_effort_parse_drawing_tool_tool_calls(
                message.tool_call_chunks
            )
    else:
        drawing_tool_chunk = _best_effort_parse_drawing_tool_tool_calls(tool_calls)

    if drawing_tool_chunk and len(drawing_tool_chunk) > 1:
        drawing_tool_result_stack = _paser_drawing_tool_chunk_input(
            message, drawing_tool_chunk
        )

    web_browser_chunk: List[
        Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]
    ] = []
    if message.tool_calls:
        if isinstance(message, ALLToolsMessageChunk):
            web_browser_chunk = _best_effort_parse_web_browser_tool_calls(
                message.tool_call_chunks
            )
    else:
        web_browser_chunk = _best_effort_parse_web_browser_tool_calls(tool_calls)

    if web_browser_chunk and len(web_browser_chunk) > 1:
        web_browser_action_result_stack = _paser_web_browser_chunk_input(
            message, web_browser_chunk
        )

    # TODO: parse platform tools built-in @langchain_glm
    # delete AdapterAllToolStructType from tool_calls
    function_tool_chunk = _best_effort_parse_function_tool_calls(tool_calls)

    function_tool_result_stack = _paser_function_chunk_input(
        message, function_tool_chunk
    )

    if isinstance(message, ALLToolsMessageChunk):
        call_chunks = _paser_object_positions(message.tool_call_chunks)

        for too_call_name in call_chunks:
            if too_call_name == AdapterAllToolStructType.CODE_INTERPRETER:
                actions.append(code_interpreter_action_result_stack.popleft())
            elif too_call_name == AdapterAllToolStructType.WEB_BROWSER:
                actions.append(web_browser_action_result_stack.popleft())
            elif too_call_name == AdapterAllToolStructType.DRAWING_TOOL:
                actions.append(drawing_tool_result_stack.popleft())
            else:
                actions.append(function_tool_result_stack.popleft())
    else:
        for too_call in tool_calls:
            if too_call["name"] not in AdapterAllToolStructType.__members__.values():
                actions.append(function_tool_result_stack.popleft())
            elif too_call["name"] == AdapterAllToolStructType.CODE_INTERPRETER:
                actions.append(code_interpreter_action_result_stack.popleft())
            elif too_call["name"] == AdapterAllToolStructType.WEB_BROWSER:
                actions.append(web_browser_action_result_stack.popleft())
            elif too_call["name"] == AdapterAllToolStructType.DRAWING_TOOL:
                actions.append(drawing_tool_result_stack.popleft())

    return actions


def _paser_object_positions(tool_call_chunks: List[ToolCallChunk]):
    call_chunks = []
    last_name = None
    if not tool_call_chunks:
        return call_chunks
    for call_chunk in tool_call_chunks:
        if call_chunk["name"] in AdapterAllToolStructType.__members__.values():
            if isinstance(call_chunk["args"], str):
                args_ = parse_partial_json(call_chunk["args"])
            else:
                args_ = call_chunk["args"]
            if not isinstance(args_, dict):
                raise ValueError("Malformed args.")

            if "outputs" in args_:
                call_chunks.append(call_chunk["name"])
                last_name = call_chunk["name"]

        else:
            if call_chunk["name"] != last_name:
                call_chunks.append(call_chunk["name"])
                last_name = call_chunk["name"]

    if len(call_chunks) == 0:
        call_chunks.append(tool_call_chunks[-1]["name"])
    elif tool_call_chunks[-1]["name"] != call_chunks[-1]:
        call_chunks.append(tool_call_chunks[-1]["name"])
    return call_chunks



# -*- coding: utf-8 -*-
import json
from typing import Any, Dict, List, Literal, Union

from langchain_core.messages import AIMessage
from langchain_core.messages.base import (
    BaseMessage,
    BaseMessageChunk,
    merge_content,
)
from langchain_core.messages.tool import (
    InvalidToolCall,
    ToolCall,
    ToolCallChunk,
    default_tool_chunk_parser,
    default_tool_parser,
    tool_call_chunk,
)
from langchain_core.pydantic_v1 import root_validator
from langchain_core.utils._merge import merge_dicts, merge_lists
from langchain_core.utils.json import (
    parse_partial_json,
)


def default_all_tool_chunk_parser(raw_tool_calls: List[dict]) -> List[ToolCallChunk]:
    """Best-effort parsing of all tool chunks."""
    tool_call_chunks = []
    for tool_call in raw_tool_calls:
        if "function" in tool_call and tool_call["function"] is not None:
            function_args = tool_call["function"]["arguments"]
            function_name = tool_call["function"]["name"]
        elif (
            "code_interpreter" in tool_call
            and tool_call["code_interpreter"] is not None
        ):
            function_args = json.dumps(
                tool_call["code_interpreter"], ensure_ascii=False
            )
            function_name = "code_interpreter"
        elif "drawing_tool" in tool_call and tool_call["drawing_tool"] is not None:
            function_args = json.dumps(tool_call["drawing_tool"], ensure_ascii=False)
            function_name = "drawing_tool"
        elif "web_browser" in tool_call and tool_call["web_browser"] is not None:
            function_args = json.dumps(tool_call["web_browser"], ensure_ascii=False)
            function_name = "web_browser"
        else:
            function_args = None
            function_name = None
        parsed = ToolCallChunk(
            name=function_name,
            args=function_args,
            id=tool_call.get("id"),
            index=tool_call.get("index"),
        )
        tool_call_chunks.append(parsed)
    return tool_call_chunks


class ALLToolsMessageChunk(AIMessage, BaseMessageChunk):
    """Message chunk from an AI."""

    # Ignoring mypy re-assignment here since we're overriding the value
    # to make sure that the chunk variant can be discriminated from the
    # non-chunk variant.
    type: Literal["ALLToolsMessageChunk"] = "ALLToolsMessageChunk"  # type: ignore[assignment] # noqa: E501

    tool_call_chunks: List[ToolCallChunk] = []
    """If provided, tool call chunks associated with the message."""

    @classmethod
    def get_lc_namespace(cls) -> List[str]:
        """Get the namespace of the langchain object."""
        return ["langchain", "schema", "messages"]

    @property
    def lc_attributes(self) -> Dict:
        """Attrs to be serialized even if they are derived from other init args."""
        return {
            "tool_calls": self.tool_calls,
            "invalid_tool_calls": self.invalid_tool_calls,
        }

    @root_validator(allow_reuse=True)
    def _backwards_compat_tool_calls(cls, values: dict) -> dict:
        raw_tool_calls = values.get("additional_kwargs", {}).get("tool_calls")
        tool_calls = (
            values.get("tool_calls")
            or values.get("invalid_tool_calls")
            or values.get("tool_call_chunks")
        )
        if raw_tool_calls and not tool_calls:
            try:
                if issubclass(cls, BaseMessageChunk):  # type: ignore
                    values["tool_call_chunks"] = default_all_tool_chunk_parser(
                        raw_tool_calls
                    )
                else:
                    tool_calls, invalid_tool_calls = default_tool_parser(raw_tool_calls)
                    values["tool_calls"] = tool_calls
                    values["invalid_tool_calls"] = invalid_tool_calls
            except Exception as e:
                pass
        return values

    @root_validator(allow_reuse=True)
    def init_tool_calls(cls, values: dict) -> dict:
        if not values["tool_call_chunks"]:
            if values["tool_calls"]:
                values["tool_call_chunks"] = [
                    tool_call_chunk(
                        name=tc["name"],
                        args=json.dumps(tc["args"]),
                        id=tc["id"],
                        index=None,
                    )
                    for tc in values["tool_calls"]
                ]
            if values["invalid_tool_calls"]:
                tool_call_chunks = values.get("tool_call_chunks", [])
                tool_call_chunks.extend(
                    [
                        tool_call_chunk(
                            name=tc["name"], args=tc["args"], id=tc["id"], index=None
                        )
                        for tc in values["invalid_tool_calls"]
                    ]
                )
                values["tool_call_chunks"] = tool_call_chunks

            return values

        tool_calls, invalid_tool_calls = _paser_chunk(values["tool_call_chunks"])
        values["tool_calls"] = tool_calls
        values["invalid_tool_calls"] = invalid_tool_calls
        return values

    def __add__(self, other: Any) -> BaseMessageChunk:  # type: ignore
        if isinstance(other, ALLToolsMessageChunk):
            if self.example != other.example:
                raise ValueError(
                    "Cannot concatenate ALLToolsMessageChunks with different example values."
                )

            content = merge_content(self.content, other.content)
            additional_kwargs = merge_dicts(
                self.additional_kwargs, other.additional_kwargs
            )
            response_metadata = merge_dicts(
                self.response_metadata, other.response_metadata
            )

            # Merge tool call chunks
            if self.tool_call_chunks or other.tool_call_chunks:
                raw_tool_calls = merge_lists(
                    self.tool_call_chunks,
                    other.tool_call_chunks,
                )
                if raw_tool_calls:
                    tool_call_chunks = [
                        ToolCallChunk(
                            name=rtc.get("name"),
                            args=rtc.get("args"),
                            index=rtc.get("index"),
                            id=rtc.get("id"),
                        )
                        for rtc in raw_tool_calls
                    ]
                else:
                    tool_call_chunks = []
            else:
                tool_call_chunks = []

            return self.__class__(
                example=self.example,
                content=content,
                additional_kwargs=additional_kwargs,
                tool_call_chunks=tool_call_chunks,
                response_metadata=response_metadata,
                id=self.id,
            )

        return super().__add__(other)


def _paser_chunk(tool_call_chunks):
    tool_calls = []
    invalid_tool_calls = []
    for chunk in tool_call_chunks:
        try:
            if "code_interpreter" in chunk["name"]:
                args_ = parse_partial_json(chunk["args"])

                if not isinstance(args_, dict):
                    raise ValueError("Malformed args.")

                if "outputs" in args_:
                    tool_calls.append(
                        ToolCall(
                            name=chunk["name"] or "",
                            args=args_,
                            id=chunk["id"],
                        )
                    )

                else:
                    invalid_tool_calls.append(
                        InvalidToolCall(
                            name=chunk["name"],
                            args=chunk["args"],
                            id=chunk["id"],
                            error=None,
                        )
                    )
            elif "drawing_tool" in chunk["name"]:
                args_ = parse_partial_json(chunk["args"])

                if not isinstance(args_, dict):
                    raise ValueError("Malformed args.")

                if "outputs" in args_:
                    tool_calls.append(
                        ToolCall(
                            name=chunk["name"] or "",
                            args=args_,
                            id=chunk["id"],
                        )
                    )

                else:
                    invalid_tool_calls.append(
                        InvalidToolCall(
                            name=chunk["name"],
                            args=chunk["args"],
                            id=chunk["id"],
                            error=None,
                        )
                    )
            elif "web_browser" in chunk["name"]:
                args_ = parse_partial_json(chunk["args"])

                if not isinstance(args_, dict):
                    raise ValueError("Malformed args.")

                if "outputs" in args_:
                    tool_calls.append(
                        ToolCall(
                            name=chunk["name"] or "",
                            args=args_,
                            id=chunk["id"],
                        )
                    )

                else:
                    invalid_tool_calls.append(
                        InvalidToolCall(
                            name=chunk["name"],
                            args=chunk["args"],
                            id=chunk["id"],
                            error=None,
                        )
                    )
            else:
                args_ = parse_partial_json(chunk["args"])

                if isinstance(args_, dict):
                    temp_args_ = {}
                    for key, value in args_.items():
                        key = key.strip()
                        temp_args_[key] = value

                    tool_calls.append(
                        ToolCall(
                            name=chunk["name"] or "",
                            args=temp_args_,
                            id=chunk["id"],
                        )
                    )
                else:
                    raise ValueError("Malformed args.")
        except Exception:
            invalid_tool_calls.append(
                InvalidToolCall(
                    name=chunk["name"],
                    args=chunk["args"],
                    id=chunk["id"],
                    error=None,
                )
            )
    return tool_calls, invalid_tool_calls



# -*- coding: utf-8 -*-
# ruff: noqa: E402
"""Main entrypoint into package."""
from importlib import metadata

from langchain_glm.agents import ZhipuAIAllToolsRunnable
from langchain_glm.chat_models import ChatZhipuAI

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""
del metadata  # optional, avoids polluting the results of dir(__package__)


__all__ = [
    "ChatZhipuAI",
    "ZhipuAIAllToolsRunnable",
]



# -*- coding: utf-8 -*-
# Function to find positions of object() instances
def find_object_positions(log_chunk, obj):
    return [i for i, x in enumerate(log_chunk) if x == obj]


# Function to concatenate segments based on object positions
def concatenate_segments(log_chunk, positions):
    segments = []
    start = 0
    for pos in positions:
        segments.append("".join(map(str, log_chunk[start:pos])))
        start = pos + 1
    return segments



# -*- coding: utf-8 -*-
import json
import logging
from collections import deque
from typing import Any, Deque, Dict, List, Union

from langchain.agents.output_parsers.tools import ToolAgentAction
from langchain_core.agents import AgentAction, AgentActionMessageLog, AgentFinish
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ToolCall,
)
from langchain_core.utils.json import (
    parse_partial_json,
)
from zhipuai.core import BaseModel

from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agents.output_parsers._utils import (
    concatenate_segments,
    find_object_positions,
)
from langchain_glm.agents.output_parsers.base import (
    AllToolsMessageToolCall,
    AllToolsMessageToolCallChunk,
)

logger = logging.getLogger(__name__)


class CodeInterpreterAgentAction(ToolAgentAction):
    outputs: List[Union[str, dict]] = None
    """Output of the tool call."""
    platform_params: dict = None


def _best_effort_parse_code_interpreter_tool_calls(
    tool_call_chunks: List[dict],
) -> List[Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]]:
    code_interpreter_chunk: List[
        Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]
    ] = []
    # Best-effort parsing allready parsed tool calls
    for code_interpreter in tool_call_chunks:
        if AdapterAllToolStructType.CODE_INTERPRETER == code_interpreter["name"]:
            if isinstance(code_interpreter["args"], str):
                args_ = parse_partial_json(code_interpreter["args"])
            else:
                args_ = code_interpreter["args"]
            if not isinstance(args_, dict):
                raise ValueError("Malformed args.")

            if "outputs" in args_:
                code_interpreter_chunk.append(
                    AllToolsMessageToolCall(
                        name=code_interpreter["name"],
                        args=args_,
                        id=code_interpreter["id"],
                    )
                )
            else:
                code_interpreter_chunk.append(
                    AllToolsMessageToolCallChunk(
                        name=code_interpreter["name"],
                        args=args_,
                        id=code_interpreter["id"],
                        index=code_interpreter.get("index"),
                    )
                )

    return code_interpreter_chunk


def _paser_code_interpreter_chunk_input(
    message: BaseMessage,
    code_interpreter_chunk: List[
        Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]
    ],
) -> Deque[CodeInterpreterAgentAction]:
    try:
        input_log_chunk = []

        outputs: List[List[dict]] = []
        obj = object()
        for interpreter_chunk in code_interpreter_chunk:
            interpreter_chunk_args = interpreter_chunk.args

            if "input" in interpreter_chunk_args:
                input_log_chunk.append(interpreter_chunk_args["input"])
            if "outputs" in interpreter_chunk_args:
                input_log_chunk.append(obj)
                outputs.append(interpreter_chunk_args["outputs"])

        if input_log_chunk[-1] is not obj:
            input_log_chunk.append(obj)
        # segments the list based on these positions, and then concatenates each segment into a string
        # Find positions of object() instances
        positions = find_object_positions(input_log_chunk, obj)

        # Concatenate segments
        result_actions = concatenate_segments(input_log_chunk, positions)

        tool_call_id = (
            code_interpreter_chunk[0].id if code_interpreter_chunk[0].id else "abc"
        )
        code_interpreter_action_result_stack: Deque[
            CodeInterpreterAgentAction
        ] = deque()
        for i, action in enumerate(result_actions):
            if len(result_actions) > len(outputs):
                outputs.insert(i, [])

            out_logs = [logs["logs"] for logs in outputs[i] if "logs" in logs]
            out_str = "\n".join(out_logs)
            log = f"{action}\r\n{out_str}"
            code_interpreter_action = CodeInterpreterAgentAction(
                tool=AdapterAllToolStructType.CODE_INTERPRETER,
                tool_input=action,
                outputs=outputs[i],
                log=log,
                message_log=[message],
                tool_call_id=tool_call_id,
            )

            code_interpreter_action_result_stack.append(code_interpreter_action)
        return code_interpreter_action_result_stack

    except Exception as e:
        logger.error(f"Error parsing code_interpreter_chunk: {e}", exc_info=True)
        raise OutputParserException(
            f"Could not parse tool input: code_interpreter because {e}"
        )



# -*- coding: utf-8 -*-
from typing import Any, Dict, Optional

from zhipuai.core import BaseModel


class AllToolsMessageToolCall(BaseModel):
    name: Optional[str]
    args: Optional[Dict[str, Any]]
    id: Optional[str]


class AllToolsMessageToolCallChunk(BaseModel):
    name: Optional[str]
    args: Optional[Dict[str, Any]]
    id: Optional[str]
    index: Optional[int]



# -*- coding: utf-8 -*-
"""OpenAI chat wrapper."""

from __future__ import annotations

import logging
import os
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    Union,
    cast,
)

import zhipuai
from langchain_core.callbacks import (
    AsyncCallbackManager,
    AsyncCallbackManagerForLLMRun,
    BaseCallbackManager,
    CallbackManager,
    CallbackManagerForLLMRun,
    Callbacks,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.language_models.chat_models import (
    BaseChatModel,
    agenerate_from_stream,
    generate_from_stream,
)
from langchain_core.load import dumpd, dumps
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    ChatMessage,
    ChatMessageChunk,
    FunctionMessage,
    FunctionMessageChunk,
    HumanMessage,
    HumanMessageChunk,
    SystemMessage,
    SystemMessageChunk,
    ToolCall,
    ToolMessage,
    ToolMessageChunk,
)
from langchain_core.outputs import (
    ChatGeneration,
    ChatGenerationChunk,
    ChatResult,
    LLMResult,
    RunInfo,
)
from langchain_core.pydantic_v1 import Field, SecretStr, root_validator
from langchain_core.runnables.config import ensure_config, run_in_executor
from langchain_core.utils import (
    convert_to_secret_str,
    get_from_dict_or_env,
    get_pydantic_field_names,
)
from langchain_core.utils.function_calling import (
    convert_to_openai_function,
    convert_to_openai_tool,
)
from langchain_core.utils.json import parse_partial_json
from langchain_core.utils.utils import build_extra_kwargs
from typing_extensions import ClassVar
from zhipuai.core import PYDANTIC_V2, ConfigDict

from langchain_glm.chat_models.all_tools_message import (
    ALLToolsMessageChunk,
    _paser_chunk,
)

if TYPE_CHECKING:
    from langchain_core.runnables import Runnable, RunnableConfig
    from langchain_core.tools import BaseTool
    from zhipuai.core import BaseModel

logger = logging.getLogger(__name__)


def _convert_dict_to_message(_dict: Mapping[str, Any]) -> BaseMessage:
    """Convert a dictionary to a LangChain message.

    Args:
        _dict: The dictionary.

    Returns:
        The LangChain message.
    """
    role = _dict.get("role")
    if role == "user":
        return HumanMessage(content=_dict.get("content", ""))
    elif role == "assistant":
        # Fix for azure
        # Also OpenAI returns None for tool invocations
        content = _dict.get("content", "") or ""
        additional_kwargs: Dict = {}
        if function_call := _dict.get("function_call"):
            additional_kwargs["function_call"] = dict(function_call)
        if tool_calls := _dict.get("tool_calls"):
            additional_kwargs["tool_calls"] = tool_calls
        return AIMessage(content=content, additional_kwargs=additional_kwargs)
    elif role == "system":
        return SystemMessage(content=_dict.get("content", ""))
    elif role == "function":
        return FunctionMessage(content=_dict.get("content", ""), name=_dict.get("name"))
    elif role == "tool":
        additional_kwargs = {}
        if "name" in _dict:
            additional_kwargs["name"] = _dict["name"]
        return ToolMessage(
            content=_dict.get("content", ""),
            tool_call_id=_dict.get("tool_call_id"),
            additional_kwargs=additional_kwargs,
        )
    else:
        return ChatMessage(content=_dict.get("content", ""), role=role)


def _convert_message_to_dict(message: BaseMessage) -> dict:
    """Convert a LangChain message to a dictionary.

    Args:
        message: The LangChain message.

    Returns:
        The dictionary.
    """
    message_dict: Dict[str, Any]
    if isinstance(message, ChatMessage):
        message_dict = {"role": message.role, "content": message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {"role": "user", "content": message.content}
    elif isinstance(message, AIMessage):
        message_dict = {"role": "assistant", "content": message.content}
        if "function_call" in message.additional_kwargs:
            message_dict["function_call"] = message.additional_kwargs["function_call"]
            # If function call only, content is None not empty string
            if message_dict["content"] == "":
                message_dict["content"] = None
        if "tool_calls" in message.additional_kwargs:
            message_dict["tool_calls"] = message.additional_kwargs["tool_calls"]
            # If tool calls only, content is None not empty string
            if message_dict["content"] == "":
                message_dict["content"] = None
    elif isinstance(message, SystemMessage):
        message_dict = {"role": "system", "content": message.content}
    elif isinstance(message, FunctionMessage):
        message_dict = {
            "role": "function",
            "content": message.content,
            "name": message.name,
        }
    elif isinstance(message, ToolMessage):
        message_dict = {
            "role": "tool",
            "content": message.content,
            "tool_call_id": message.tool_call_id,
        }
    else:
        raise TypeError(f"Got unknown type {message}")
    if "name" in message.additional_kwargs:
        message_dict["name"] = message.additional_kwargs["name"]
    return message_dict


def _convert_delta_to_message_chunk(
    _dict: Mapping[str, Any], default_class: Type[BaseMessageChunk]
) -> BaseMessageChunk:
    role = cast(str, _dict.get("role"))
    content = cast(str, _dict.get("content") or "")
    additional_kwargs: Dict = {}
    if _dict.get("function_call"):
        function_call = dict(_dict["function_call"])
        if "name" in function_call and function_call["name"] is None:
            function_call["name"] = ""
        additional_kwargs["function_call"] = function_call
    if _dict.get("tool_calls"):
        additional_kwargs["tool_calls"] = _dict["tool_calls"]

    if role == "user" or default_class == HumanMessageChunk:
        return HumanMessageChunk(content=content)
    elif default_class == ALLToolsMessageChunk:
        return ALLToolsMessageChunk(
            content=content, additional_kwargs=additional_kwargs
        )
    elif role == "assistant" or default_class == AIMessageChunk:
        return AIMessageChunk(content=content, additional_kwargs=additional_kwargs)
    elif role == "system" or default_class == SystemMessageChunk:
        return SystemMessageChunk(content=content)
    elif role == "function" or default_class == FunctionMessageChunk:
        return FunctionMessageChunk(content=content, name=_dict["name"])
    elif role == "tool" or default_class == ToolMessageChunk:
        return ToolMessageChunk(content=content, tool_call_id=_dict["tool_call_id"])
    elif role or default_class == ChatMessageChunk:
        return ChatMessageChunk(content=content, role=role)
    else:
        return default_class(content=content)  # type: ignore


class _FunctionCall(TypedDict):
    name: str


class ChatZhipuAI(BaseChatModel):
    """ZhipuAI chat model integration.

    Setup:
        Install ``langchain-zhipuai`` and set environment variable ``ZHIPUAI_API_KEY``

        .. code-block:: bash

            pip install langchain-zhipuai -U
            export ZHIPUAI_API_KEY="your-api-key"

    Key init args — completion params:
        model: Optional[str]
            Name of OpenAI model to use.
        temperature: float
            Sampling temperature.
        max_tokens: Optional[int]
            Max number of tokens to generate.

    Key init args — client params:
        api_key: Optional[str]
        ZhipuAI API key. If not passed in will be read from env var ZHIPUAI_API_KEY.
        api_base: Optional[str]
        Base URL for API requests.

    See full list of supported init args and their descriptions in the params section.

    Instantiate:
        .. code-block:: python

            from langchain_community.chat_models import ChatZhipuAI

            zhipuai_chat = ChatZhipuAI(
                temperature=0.5,
                api_key="your-api-key",
                model="glm-4",
                # api_base="...",
                # other params...
            )

    Invoke:
        .. code-block:: python

            messages = [
                ("system", "你是一名专业的翻译家，可以将用户的中文翻译为英文。"),
                ("human", "我喜欢编程。"),
            ]
            zhipuai_chat.invoke(messages)

        .. code-block:: python

            AIMessage(content='I enjoy programming.', response_metadata={'token_usage': {'completion_tokens': 6, 'prompt_tokens': 23, 'total_tokens': 29}, 'model_name': 'glm-4', 'finish_reason': 'stop'}, id='run-c5d9af91-55c6-470e-9545-02b2fa0d7f9d-0')

    Stream:
        .. code-block:: python

            for chunk in zhipuai_chat.stream(messages):
                print(chunk)

        .. code-block:: python

            content='I' id='run-4df71729-618f-4e2b-a4ff-884682723082'
            content=' enjoy' id='run-4df71729-618f-4e2b-a4ff-884682723082'
            content=' programming' id='run-4df71729-618f-4e2b-a4ff-884682723082'
            content='.' id='run-4df71729-618f-4e2b-a4ff-884682723082'
            content='' response_metadata={'finish_reason': 'stop'} id='run-4df71729-618f-4e2b-a4ff-884682723082'

        .. code-block:: python

            stream = llm.stream(messages)
            full = next(stream)
            for chunk in stream:
                full += chunk
            full

        .. code-block::

            AIMessageChunk(content='I enjoy programming.', response_metadata={'finish_reason': 'stop'}, id='run-20b05040-a0b4-4715-8fdc-b39dba9bfb53')

    Async:
        .. code-block:: python

            await zhipuai_chat.ainvoke(messages)

            # stream:
            # async for chunk in zhipuai_chat.astream(messages):
            #    print(chunk)


        .. code-block:: python

            [AIMessage(content='I enjoy programming.', response_metadata={'token_usage': {'completion_tokens': 6, 'prompt_tokens': 23, 'total_tokens': 29}, 'model_name': 'glm-4', 'finish_reason': 'stop'}, id='run-ba06af9d-4baa-40b2-9298-be9c62aa0849-0')]

    Response metadata
        .. code-block:: python

            ai_msg = zhipuai_chat.invoke(messages)
            ai_msg.response_metadata

        .. code-block:: python

            {'token_usage': {'completion_tokens': 6,
              'prompt_tokens': 23,
              'total_tokens': 29},
              'model_name': 'glm-4',
              'finish_reason': 'stop'}

    """  # noqa: E501

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"zhipuai_api_key": "ZHIPUAI_API_KEY"}

    @classmethod
    def get_lc_namespace(cls) -> List[str]:
        """Get the namespace of the langchain object."""
        return ["langchain", "chat_models", "openai"]

    @property
    def lc_attributes(self) -> Dict[str, Any]:
        attributes: Dict[str, Any] = {}

        if self.zhipuai_api_base:
            attributes["zhipuai_api_base"] = self.zhipuai_api_base

        if self.zhipuai_proxy:
            attributes["zhipuai_proxy"] = self.zhipuai_proxy

        return attributes

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Return whether this model can be serialized by Langchain."""
        return True

    client: Any = Field(default=None, exclude=True)  #: :meta private:
    model_name: str = Field(default="glm-4", alias="model")
    """Model name to use."""
    temperature: float = 0.7
    """What sampling temperature to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    zhipuai_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    """Automatically inferred from env var `ZHIPUAI_API_KEY` if not provided."""
    zhipuai_api_base: Optional[str] = Field(default=None, alias="base_url")
    """Base URL path for API requests, leave blank if not using a proxy or service 
        emulator."""
    # to support explicit proxy for OpenAI
    zhipuai_proxy: Optional[str] = None
    request_timeout: Union[float, Tuple[float, float], Any, None] = Field(
        default=None, alias="timeout"
    )
    """Timeout for requests to OpenAI completion API. Can be float, httpx.Timeout or 
        None."""
    max_retries: int = 2
    """Maximum number of retries to make when generating."""
    streaming: bool = False
    """Whether to stream the results or not."""
    max_tokens: Optional[int] = None
    """Maximum number of tokens to generate."""
    http_client: Union[Any, None] = None
    """Optional httpx.Client."""

    if PYDANTIC_V2:
        model_config: ClassVar[ConfigDict] = ConfigDict(populate_by_name=True)
    else:

        class Config:
            allow_population_by_field_name = True

    @root_validator(pre=True, allow_reuse=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)
        extra = values.get("model_kwargs", {})
        values["model_kwargs"] = build_extra_kwargs(
            extra, values, all_required_field_names
        )
        return values

    @root_validator(allow_reuse=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""

        values["zhipuai_api_key"] = convert_to_secret_str(
            get_from_dict_or_env(values, "zhipuai_api_key", "ZHIPUAI_API_KEY")
        )

        values["zhipuai_api_base"] = values["zhipuai_api_base"] or os.getenv(
            "ZHIPUAI_API_BASE"
        )
        values["zhipuai_proxy"] = get_from_dict_or_env(
            values,
            "zhipuai_proxy",
            "ZHIPUAI_PROXY",
            default="",
        )

        client_params = {
            "api_key": (
                values["zhipuai_api_key"].get_secret_value()
                if values["zhipuai_api_key"]
                else None
            ),
            "base_url": values["zhipuai_api_base"],
            "timeout": values["request_timeout"],
            "max_retries": values["max_retries"],
            "http_client": values["http_client"],
        }

        if not values.get("client"):
            values["client"] = zhipuai.ZhipuAI(**client_params).chat.completions

        return values

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        params = {
            "model": self.model_name,
            "stream": self.streaming,
            "temperature": self.temperature,
            **self.model_kwargs,
        }
        if self.max_tokens is not None:
            params["max_tokens"] = self.max_tokens
        return params

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        overall_token_usage: dict = {}
        system_fingerprint = None
        for output in llm_outputs:
            if output is None:
                # Happens in streaming
                continue
            token_usage = output["token_usage"]
            if token_usage is not None:
                for k, v in token_usage.items():
                    if k in overall_token_usage:
                        overall_token_usage[k] += v
                    else:
                        overall_token_usage[k] = v
            if system_fingerprint is None:
                system_fingerprint = output.get("system_fingerprint")
        combined = {"token_usage": overall_token_usage, "model_name": self.model_name}
        if system_fingerprint:
            combined["system_fingerprint"] = system_fingerprint
        return combined

    def stream(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        *,
        stop: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Iterator[BaseMessageChunk]:
        if type(self)._stream == BaseChatModel._stream:
            # model doesn't implement streaming, so use default implementation
            yield cast(
                BaseMessageChunk, self.invoke(input, config=config, stop=stop, **kwargs)
            )
        else:
            config = ensure_config(config)
            messages = self._convert_input(input).to_messages()
            params = self._get_invocation_params(stop=stop, **kwargs)
            options = {"stop": stop, **kwargs}
            callback_manager = CallbackManager.configure(
                config.get("callbacks"),
                self.callbacks,
                self.verbose,
                config.get("tags"),
                self.tags,
                config.get("metadata"),
                self.metadata,
            )
            (run_manager,) = callback_manager.on_chat_model_start(
                dumpd(self),
                [messages],
                invocation_params=params,
                options=options,
                name=config.get("run_name"),
                run_id=config.pop("run_id", None),
                batch_size=1,
            )
            generation: Optional[ChatGenerationChunk] = None
            try:
                for chunk in self._stream(messages, stop=stop, **kwargs):
                    if chunk.message.id is None:
                        chunk.message.id = f"run-{run_manager.run_id}"
                    chunk.message.response_metadata = _gen_info_and_msg_metadata(chunk)
                    if (
                        isinstance(chunk.message, ALLToolsMessageChunk)
                        and chunk.message.content == ""
                    ):
                        tool_calls, invalid_tool_calls = _paser_chunk(
                            chunk.message.tool_call_chunks
                        )

                        for chunk_tool in invalid_tool_calls:
                            if isinstance(chunk_tool["args"], str):
                                args_ = parse_partial_json(chunk_tool["args"])
                            else:
                                args_ = chunk_tool["args"]
                            if not isinstance(args_, dict):
                                raise ValueError("Malformed args.")
                            if "input" in args_:
                                run_manager.on_llm_new_token(
                                    cast(str, args_["input"]), chunk=chunk
                                )

                    else:
                        run_manager.on_llm_new_token(
                            cast(str, chunk.message.content), chunk=chunk
                        )
                    yield chunk.message
                    if generation is None:
                        generation = chunk
                    else:
                        generation += chunk
                assert generation is not None
            except BaseException as e:
                run_manager.on_llm_error(
                    e,
                    response=LLMResult(
                        generations=[[generation]] if generation else []
                    ),
                )
                raise e
            else:
                run_manager.on_llm_end(LLMResult(generations=[[generation]]))

    async def astream(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        *,
        stop: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[BaseMessageChunk]:
        if (
            type(self)._astream is BaseChatModel._astream
            and type(self)._stream is BaseChatModel._stream
        ):
            # No async or sync stream is implemented, so fall back to ainvoke
            yield cast(
                BaseMessageChunk,
                await self.ainvoke(input, config=config, stop=stop, **kwargs),
            )
            return

        config = ensure_config(config)
        messages = self._convert_input(input).to_messages()
        params = self._get_invocation_params(stop=stop, **kwargs)
        options = {"stop": stop, **kwargs}
        callback_manager = AsyncCallbackManager.configure(
            config.get("callbacks"),
            self.callbacks,
            self.verbose,
            config.get("tags"),
            self.tags,
            config.get("metadata"),
            self.metadata,
        )
        (run_manager,) = await callback_manager.on_chat_model_start(
            dumpd(self),
            [messages],
            invocation_params=params,
            options=options,
            name=config.get("run_name"),
            run_id=config.pop("run_id", None),
            batch_size=1,
        )

        generation: Optional[ChatGenerationChunk] = None
        try:
            async for chunk in self._astream(
                messages,
                stop=stop,
                **kwargs,
            ):
                if chunk.message.id is None:
                    chunk.message.id = f"run-{run_manager.run_id}"
                chunk.message.response_metadata = _gen_info_and_msg_metadata(chunk)
                if (
                    isinstance(chunk.message, ALLToolsMessageChunk)
                    and chunk.message.content == ""
                ):
                    tool_calls, invalid_tool_calls = _paser_chunk(
                        chunk.message.tool_call_chunks
                    )

                    for chunk_tool in invalid_tool_calls:
                        if isinstance(chunk_tool["args"], str):
                            try:
                                args_ = parse_partial_json(chunk_tool["args"])
                            except Exception as e:
                                args_ = {"input": chunk_tool["args"]}
                        else:
                            args_ = chunk_tool["args"]
                        if not isinstance(args_, dict):
                            raise ValueError("Malformed args.")
                        if "input" in args_:
                            await run_manager.on_llm_new_token(
                                cast(str, args_["input"]), chunk=chunk
                            )
                        else:
                            await run_manager.on_llm_new_token(
                                cast(str, args_), chunk=chunk
                            )
                else:
                    await run_manager.on_llm_new_token(
                        cast(str, chunk.message.content), chunk=chunk
                    )
                yield chunk.message
                if generation is None:
                    generation = chunk
                else:
                    generation += chunk
            assert generation is not None
        except BaseException as e:
            await run_manager.on_llm_error(
                e,
                response=LLMResult(generations=[[generation]] if generation else []),
            )
            raise e
        else:
            await run_manager.on_llm_end(
                LLMResult(generations=[[generation]]),
            )

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        for chunk in self.client.create(messages=message_dicts, **params):
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            # all_tools chunk load action exec parse tool
            if params["model"] in [
                "glm-4-alltools-dev",
                "tob-alltools-api-dev",
                "glm-4-alltools",
            ]:
                default_chunk_class = ALLToolsMessageChunk

            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            generation_info = {}
            if finish_reason := choice.get("finish_reason"):
                generation_info["finish_reason"] = finish_reason
            logprobs = choice.get("logprobs")
            if logprobs:
                generation_info["logprobs"] = logprobs
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(
                message=chunk, generation_info=generation_info or None
            )
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk, logprobs=logprobs)
            yield chunk

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        stream: Optional[bool] = None,
        **kwargs: Any,
    ) -> ChatResult:
        should_stream = stream if stream is not None else self.streaming
        if should_stream:
            stream_iter = self._stream(
                messages, stop=stop, run_manager=run_manager, **kwargs
            )
            return generate_from_stream(stream_iter)
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {
            **params,
            **({"stream": stream} if stream is not None else {}),
            **kwargs,
        }
        response = self.client.create(messages=message_dicts, **params)
        return self._create_chat_result(response)

    def _create_message_dicts(
        self, messages: List[BaseMessage], stop: Optional[List[str]]
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        params = self._default_params
        if stop is not None:
            if "stop" in params:
                raise ValueError("`stop` found in both the input and default params.")
            params["stop"] = stop
        message_dicts = [_convert_message_to_dict(m) for m in messages]
        return message_dicts, params

    def _create_chat_result(self, response: Union[dict, BaseModel]) -> ChatResult:
        generations = []
        if not isinstance(response, dict):
            response = response.dict()
        for res in response["choices"]:
            message = _convert_dict_to_message(res["message"])
            generation_info = dict(finish_reason=res.get("finish_reason"))
            if "logprobs" in res:
                generation_info["logprobs"] = res["logprobs"]
            gen = ChatGeneration(
                message=message,
                generation_info=generation_info,
            )
            generations.append(gen)
        token_usage = response.get("usage", {})
        llm_output = {
            "token_usage": token_usage,
            "model_name": self.model_name,
            "system_fingerprint": response.get("system_fingerprint", ""),
        }
        return ChatResult(generations=generations, llm_output=llm_output)

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {"model_name": self.model_name, **self._default_params}

    def _get_invocation_params(
        self, stop: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Any]:
        """Get the parameters used to invoke the model."""
        return {
            "model": self.model_name,
            **super()._get_invocation_params(stop=stop),
            **self._default_params,
            **kwargs,
        }

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "zhipuai-chat"

    def bind_functions(
        self,
        functions: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool]],
        function_call: Optional[
            Union[_FunctionCall, str, Literal["auto", "none"]]
        ] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        """Bind functions (and other objects) to this chat model.

        Assumes model is compatible with OpenAI function-calling API.

        NOTE: Using bind_tools is recommended instead, as the `functions` and
            `function_call` request parameters are officially marked as deprecated by
            OpenAI.

        Args:
            functions: A list of function definitions to bind to this chat model.
                Can be  a dictionary, pydantic model, or callable. Pydantic
                models and callables will be automatically converted to
                their schema dictionary representation.
            function_call: Which function to require the model to call.
                Must be the name of the single provided function or
                "auto" to automatically determine which function to call
                (if any).
            **kwargs: Any additional parameters to pass to the
                :class:`~langchain.runnable.Runnable` constructor.
        """

        formatted_functions = [convert_to_openai_function(fn) for fn in functions]
        if function_call is not None:
            function_call = (
                {"name": function_call}
                if isinstance(function_call, str)
                and function_call not in ("auto", "none")
                else function_call
            )
            if isinstance(function_call, dict) and len(formatted_functions) != 1:
                raise ValueError(
                    "When specifying `function_call`, you must provide exactly one "
                    "function."
                )
            if (
                isinstance(function_call, dict)
                and formatted_functions[0]["name"] != function_call["name"]
            ):
                raise ValueError(
                    f"Function call {function_call} was specified, but the only "
                    f"provided function was {formatted_functions[0]['name']}."
                )
            kwargs = {**kwargs, "function_call": function_call}
        return super().bind(
            functions=formatted_functions,
            **kwargs,
        )

    def bind_tools(
        self,
        tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool]],
        *,
        tool_choice: Optional[Union[dict, str, Literal["auto", "none"]]] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        """Bind tool-like objects to this chat model.

        Assumes model is compatible with OpenAI tool-calling API.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be  a dictionary, pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            tool_choice: Which tool to require the model to call.
                Must be the name of the single provided function or
                "auto" to automatically determine which function to call
                (if any), or a dict of the form:
                {"type": "function", "function": {"name": <<tool_name>>}}.
            **kwargs: Any additional parameters to pass to the
                :class:`~langchain.runnable.Runnable` constructor.
        """

        formatted_tools = [convert_to_openai_tool(tool) for tool in tools]
        if tool_choice is not None:
            if isinstance(tool_choice, str) and (tool_choice not in ("auto", "none")):
                tool_choice = {"type": "function", "function": {"name": tool_choice}}
            if isinstance(tool_choice, dict) and (len(formatted_tools) != 1):
                raise ValueError(
                    "When specifying `tool_choice`, you must provide exactly one "
                    f"tool. Received {len(formatted_tools)} tools."
                )
            if isinstance(tool_choice, dict) and (
                formatted_tools[0]["function"]["name"]
                != tool_choice["function"]["name"]
            ):
                raise ValueError(
                    f"Tool choice {tool_choice} was specified, but the only "
                    f"provided tool was {formatted_tools[0]['function']['name']}."
                )
            kwargs["tool_choice"] = tool_choice
        return super().bind(tools=formatted_tools, **kwargs)


def _gen_info_and_msg_metadata(
    generation: Union[ChatGeneration, ChatGenerationChunk],
) -> dict:
    return {
        **(generation.generation_info or {}),
        **generation.message.response_metadata,
    }



# -*- coding: utf-8 -*-
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from langchain_core.agents import AgentAction
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)

from langchain_glm.agent_toolkits import AdapterAllTool
from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agent_toolkits.all_tools.tool import (
    AllToolExecutor,
    BaseToolOutput,
)

logger = logging.getLogger(__name__)


class DrawingToolOutput(BaseToolOutput):
    platform_params: Dict[str, Any]

    def __init__(
        self,
        data: Any,
        platform_params: Dict[str, Any],
        **extras: Any,
    ) -> None:
        super().__init__(data, "", "", **extras)
        self.platform_params = platform_params


@dataclass
class DrawingAllToolExecutor(AllToolExecutor):
    """platform adapter tool for code interpreter tool"""

    name: str

    def run(
        self,
        tool: str,
        tool_input: str,
        log: str,
        outputs: List[Union[str, dict]] = None,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> DrawingToolOutput:
        if outputs is None or str(outputs).strip() == "":
            raise ValueError(f"Tool {self.name}  is server error")

        return DrawingToolOutput(
            data=f"""Access：{tool}, Message: {tool_input},{log}""",
            platform_params=self.platform_params,
        )

    async def arun(
        self,
        tool: str,
        tool_input: str,
        log: str,
        outputs: List[Union[str, dict]] = None,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> DrawingToolOutput:
        """Use the tool asynchronously."""
        if outputs is None or str(outputs).strip() == "" or len(outputs) == 0:
            raise ValueError(f"Tool {self.name}  is server error")

        return DrawingToolOutput(
            data=f"""Access：{tool}, Message: {tool_input},{log}""",
            platform_params=self.platform_params,
        )


class DrawingAdapterAllTool(AdapterAllTool[DrawingAllToolExecutor]):
    @classmethod
    def get_type(cls) -> str:
        return "DrawingAdapterAllTool"

    def _build_adapter_all_tool(
        self, platform_params: Dict[str, Any]
    ) -> DrawingAllToolExecutor:
        return DrawingAllToolExecutor(
            name=AdapterAllToolStructType.DRAWING_TOOL, platform_params=platform_params
        )



# -*- coding: utf-8 -*-
import json
import logging
from collections import deque
from typing import Any, Deque, Dict, List, Union

from langchain.agents.output_parsers.tools import ToolAgentAction
from langchain_core.agents import AgentAction, AgentActionMessageLog, AgentFinish
from langchain_core.exceptions import OutputParserException
from langchain_core.messages import (
    BaseMessage,
    ToolCall,
)
from langchain_core.utils.json import parse_partial_json

from langchain_glm.agent_toolkits.all_tools.struct_type import AdapterAllToolStructType
from langchain_glm.agents.output_parsers._utils import (
    concatenate_segments,
    find_object_positions,
)
from langchain_glm.agents.output_parsers.base import (
    AllToolsMessageToolCall,
    AllToolsMessageToolCallChunk,
)

logger = logging.getLogger(__name__)


def _best_effort_parse_function_tool_calls(
    tool_call_chunks: List[dict],
) -> List[Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]]:
    function_chunk: List[
        Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]
    ] = []
    # Best-effort parsing allready parsed tool calls
    for function in tool_call_chunks:
        if function["name"] not in AdapterAllToolStructType.__members__.values():
            if isinstance(function["args"], str):
                args_ = parse_partial_json(function["args"])
            else:
                args_ = function["args"]
            if not isinstance(args_, dict):
                raise ValueError("Malformed args.")

            if len(args_.keys()) > 0:
                function_chunk.append(
                    AllToolsMessageToolCall(
                        name=function["name"],
                        args=args_,
                        id=function["id"],
                    )
                )
            else:
                function_chunk.append(
                    AllToolsMessageToolCallChunk(
                        name=function["name"],
                        args=args_,
                        id=function["id"],
                        index=function.get("index"),
                    )
                )

    return function_chunk


def _paser_function_chunk_input(
    message: BaseMessage,
    function_chunk: List[Union[AllToolsMessageToolCall, AllToolsMessageToolCallChunk]],
) -> Deque[ToolAgentAction]:
    try:
        function_action_result_stack: Deque[ToolAgentAction] = deque()
        for _chunk in function_chunk:
            if isinstance(_chunk, AllToolsMessageToolCall):
                function_name = _chunk.name
                _tool_input = _chunk.args
                tool_call_id = _chunk.id if _chunk.id else "abc"
                if "__arg1" in _tool_input:
                    tool_input = _tool_input["__arg1"]
                else:
                    tool_input = _tool_input

                content_msg = (
                    f"responded: {message.content}\n" if message.content else "\n"
                )
                log = f"\nInvoking: `{function_name}` with `{tool_input}`\n{content_msg}\n"

                function_action_result_stack.append(
                    ToolAgentAction(
                        tool=function_name,
                        tool_input=tool_input,
                        log=log,
                        message_log=[message],
                        tool_call_id=tool_call_id,
                    )
                )

        return function_action_result_stack

    except Exception as e:
        logger.error(f"Error parsing function_chunk: {e}", exc_info=True)
        raise OutputParserException(f"Error parsing function_chunk: {e} ")



# -*- coding: utf-8 -*-
import logging
from functools import lru_cache
from typing import Any, Dict, List, Tuple, Union

from langchain.prompts.chat import ChatMessagePromptTemplate
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    ChatMessage,
    ChatMessageChunk,
    FunctionMessage,
    FunctionMessageChunk,
    HumanMessage,
    HumanMessageChunk,
    SystemMessage,
    SystemMessageChunk,
    ToolMessage,
    ToolMessageChunk,
)
from zhipuai.core import BaseModel

logger = logging.getLogger()


def _convert_message_to_dict(message: BaseMessage) -> dict:
    """Convert a LangChain message to a dictionary.

    Args:
        message: The LangChain message.

    Returns:
        The dictionary.
    """
    message_dict: Dict[str, Any]
    if isinstance(message, ChatMessage):
        message_dict = {"role": message.role, "content": message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {"role": "user", "content": message.content}
    elif isinstance(message, AIMessage):
        message_dict = {"role": "assistant", "content": message.content}
        if "function_call" in message.additional_kwargs:
            message_dict["function_call"] = message.additional_kwargs["function_call"]
            # If function call only, content is None not empty string
            if message_dict["content"] == "":
                message_dict["content"] = None
        if "tool_calls" in message.additional_kwargs:
            message_dict["tool_calls"] = message.additional_kwargs["tool_calls"]
            # If tool calls only, content is None not empty string
            if message_dict["content"] == "":
                message_dict["content"] = None
    elif isinstance(message, SystemMessage):
        message_dict = {"role": "system", "content": message.content}
    elif isinstance(message, FunctionMessage):
        message_dict = {
            "role": "function",
            "content": message.content,
            "name": message.name,
        }
    elif isinstance(message, ToolMessage):
        message_dict = {
            "role": "tool",
            "content": message.content,
            "tool_call_id": message.tool_call_id,
        }
    else:
        raise TypeError(f"Got unknown type {message}")
    if "name" in message.additional_kwargs:
        message_dict["name"] = message.additional_kwargs["name"]
    return message_dict


class History(BaseModel):
    """
    对话历史
    可从dict生成，如
    h = History(**{"role":"user","content":"你好"})
    也可转换为tuple，如
    h.to_msy_tuple = ("human", "你好")
    """

    role: str
    content: str

    def to_msg_tuple(self):
        return "ai" if self.role == "assistant" else "human", self.content

    def to_msg_template(self, is_raw=True) -> ChatMessagePromptTemplate:
        role_maps = {
            "ai": "assistant",
            "human": "user",
        }
        role = role_maps.get(self.role, self.role)
        if is_raw:  # 当前默认历史消息都是没有input_variable的文本。
            content = "{% raw %}" + self.content + "{% endraw %}"
        else:
            content = self.content

        return ChatMessagePromptTemplate.from_template(
            content,
            "jinja2",
            role=role,
        )

    @classmethod
    def from_data(cls, h: Union[List, Tuple, Dict]) -> "History":
        if isinstance(h, (list, tuple)) and len(h) >= 2:
            h = cls(role=h[0], content=h[1])
        elif isinstance(h, dict):
            h = cls(**h)

        return h

    @classmethod
    def from_message(cls, message: BaseMessage) -> "History":
        return cls.from_data(_convert_message_to_dict(message=message))



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import os
import warnings
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

import zhipuai
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import (
    BaseModel,
    Extra,
    Field,
    SecretStr,
    root_validator,
)
from langchain_core.utils import (
    convert_to_secret_str,
    get_from_dict_or_env,
    get_pydantic_field_names,
)

logger = logging.getLogger(__name__)


class ZhipuAIEmbeddings(BaseModel, Embeddings):
    """ZhipuAI embedding models.

    To use, you should have the
    environment variable ``OPENAI_API_KEY`` set with your API key or pass it
    as a named parameter to the constructor.

    Example:
        .. code-block:: python

            from langchain_glm import ZhipuAIEmbeddings

            zhipuai = ZhipuAIEmbeddings(model=""text_embedding")


    """

    client: Any = Field(default=None, exclude=True)  #: :meta private:
    model: str = "embedding-2"
    zhipuai_api_base: Optional[str] = Field(default=None, alias="base_url")
    """Base URL path for API requests, leave blank if not using a proxy or service 
        emulator."""
    zhipuai_proxy: Optional[str] = None
    embedding_ctx_length: int = 8191
    """The maximum number of tokens to embed at once."""
    zhipuai_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    """Automatically inferred from env var `OPENAI_API_KEY` if not provided."""

    chunk_size: int = 1000
    """Maximum number of texts to embed in each batch"""
    max_retries: int = 2
    """Maximum number of retries to make when generating."""
    request_timeout: Optional[Union[float, Tuple[float, float], Any]] = Field(
        default=None, alias="timeout"
    )
    """Timeout for requests to OpenAI completion API. Can be float, httpx.Timeout or 
        None."""
    headers: Any = None

    show_progress_bar: bool = False
    """Whether to show a progress bar when embedding."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    http_client: Union[Any, None] = None
    """Optional httpx.Client."""


    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        allow_population_by_field_name = True

    @root_validator(pre=True, allow_reuse=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)
        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name in extra:
                raise ValueError(f"Found {field_name} supplied twice.")
            if field_name not in all_required_field_names:
                warnings.warn(
                    f"""WARNING! {field_name} is not default parameter.
                    {field_name} was transferred to model_kwargs.
                    Please confirm that {field_name} is what you intended."""
                )
                extra[field_name] = values.pop(field_name)

        invalid_model_kwargs = all_required_field_names.intersection(extra.keys())
        if invalid_model_kwargs:
            raise ValueError(
                f"Parameters {invalid_model_kwargs} should be specified explicitly. "
                f"Instead they were passed in as part of `model_kwargs` parameter."
            )

        values["model_kwargs"] = extra
        return values

    @root_validator(allow_reuse=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        zhipuai_api_key = get_from_dict_or_env(
            values, "zhipuai_api_key", "ZHIPUAI_API_KEY"
        )
        values["zhipuai_api_key"] = (
            convert_to_secret_str(zhipuai_api_key) if zhipuai_api_key else None
        )
        values["zhipuai_api_base"] = values["zhipuai_api_base"] or os.getenv(
            "OPENAI_API_BASE"
        )
        values["zhipuai_api_type"] = get_from_dict_or_env(
            values,
            "zhipuai_api_type",
            "OPENAI_API_TYPE",
            default="",
        )
        values["zhipuai_proxy"] = get_from_dict_or_env(
            values,
            "zhipuai_proxy",
            "OPENAI_PROXY",
            default="",
        )

        client_params = {
            "api_key": values["zhipuai_api_key"].get_secret_value()
            if values["zhipuai_api_key"]
            else None,
            "base_url": values["zhipuai_api_base"],
            "timeout": values["request_timeout"],
            "max_retries": values["max_retries"],
            "http_client": values["http_client"],
        }
        if not values.get("client"):
            values["client"] = zhipuai.ZhipuAI(**client_params).embeddings
        return values

    @property
    def _invocation_params(self) -> Dict[str, Any]:
        params: Dict = {"model": self.model, **self.model_kwargs}
        return params

    def _get_len_safe_embeddings(
        self, texts: List[str], *, chunk_size: Optional[int] = None
    ) -> List[List[float]]:
        """
        Generate length-safe embeddings for a list of texts.
        Args:
            texts (List[str]): A list of texts to embed.
            chunk_size (Optional[int]): The size of chunks for processing embeddings.

        Returns:
            List[List[float]]: A list of embeddings for each input text.
        """

        _chunk_size = chunk_size or self.chunk_size

        if self.show_progress_bar:
            try:
                from tqdm.auto import tqdm

                _iter: Iterable = tqdm(range(0, len(texts), _chunk_size))
            except ImportError:
                _iter = range(0, len(texts), _chunk_size)
        else:
            _iter = range(0, len(texts), _chunk_size)

        batched_embeddings: List[List[float]] = []
        for i in _iter:
            response = self.client.create(
                input=texts[i : i + _chunk_size], **self._invocation_params
            )
            if not isinstance(response, dict):
                response = response.dict()
            batched_embeddings.extend(r["embedding"] for r in response["data"])

        return batched_embeddings

    def embed_documents(
        self, texts: List[str], chunk_size: Optional[int] = 0
    ) -> List[List[float]]:
        """Call out to OpenAI's embedding endpoint for embedding search docs.

        Args:
            texts: The list of texts to embed.
            chunk_size: The chunk size of embeddings. If None, will use the chunk size
                specified by the class.

        Returns:
            List of embeddings, one for each text.
        """
        return self._get_len_safe_embeddings(texts)

    def embed_query(self, text: str) -> List[float]:
        """Call out to OpenAI's embedding endpoint for embedding query text.

        Args:
            text: The text to embed.

        Returns:
            Embedding for the text.
        """
        return self.embed_documents([text])[0]



# -*- coding: utf-8 -*-
"""IndexStructType class."""

from enum import Enum


class AdapterAllToolStructType(str, Enum):
    """

    Attributes:
        DICT ("dict"):

    """

    # TODO: refactor so these are properties on the base class

    CODE_INTERPRETER = "code_interpreter"
    DRAWING_TOOL = "drawing_tool"
    WEB_BROWSER = "web_browser"



# -*- coding: utf-8 -*-
import asyncio
import json
import logging
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from langchain import hub
from langchain.agents import AgentExecutor
from langchain_core.agents import AgentAction
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.language_models import BaseLanguageModel
from langchain_core.messages import convert_to_messages
from langchain_core.runnables import RunnableConfig, RunnableSerializable
from langchain_core.runnables.base import RunnableBindingBase
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from pydantic.v1 import Field, validator
from typing_extensions import ClassVar
from zhipuai.core import PYDANTIC_V2, BaseModel, ConfigDict

from langchain_glm.agent_toolkits.all_tools.registry import (
    TOOL_STRUCT_TYPE_TO_TOOL_CLASS,
)
from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agent_toolkits.all_tools.tool import (
    AdapterAllTool,
    BaseToolOutput,
)
from langchain_glm.agents.all_tools_agent import ZhipuAiAllToolsAgentExecutor
from langchain_glm.agents.all_tools_bind.base import create_zhipuai_tools_agent
from langchain_glm.agents.format_scratchpad.all_tools import (
    format_to_zhipuai_all_tool_messages,
)
from langchain_glm.agents.output_parsers import ZhipuAiALLToolsAgentOutputParser
from langchain_glm.agents.zhipuai_all_tools.schema import (
    AllToolsAction,
    AllToolsActionToolEnd,
    AllToolsActionToolStart,
    AllToolsFinish,
    AllToolsLLMStatus,
)
from langchain_glm.callbacks.agent_callback_handler import (
    AgentExecutorAsyncIteratorCallbackHandler,
    AgentStatus,
)
from langchain_glm.chat_models import ChatZhipuAI
from langchain_glm.utils import History

logger = logging.getLogger()


def _is_assistants_builtin_tool(
    tool: Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool],
) -> bool:
    """platform tools built-in"""
    assistants_builtin_tools = AdapterAllToolStructType.__members__.values()
    return (
        isinstance(tool, dict)
        and ("type" in tool)
        and (tool["type"] in assistants_builtin_tools)
    )


def _get_assistants_tool(
    tool: Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool],
) -> Dict[str, Any]:
    """Convert a raw function/class to an ZhipuAI tool."""
    if _is_assistants_builtin_tool(tool):
        return tool  # type: ignore
    else:
        # in case of a custom tool, convert it to an function of type
        return convert_to_openai_tool(tool)


def _agents_registry(
    llm: BaseLanguageModel,
    llm_with_all_tools: RunnableBindingBase = None,
    tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool]] = [],
    callbacks: List[BaseCallbackHandler] = [],
    verbose: bool = False,
):
    if llm_with_all_tools:
        prompt = hub.pull("zhipuai-all-tools-chat/zhipuai-all-tools-agent")
        agent = create_zhipuai_tools_agent(
            prompt=prompt, llm_with_all_tools=llm_with_all_tools
        )
    else:
        prompt = hub.pull("zhipuai-all-tools-chat/zhipuai-all-tools-chat")
        agent = prompt | llm | ZhipuAiALLToolsAgentOutputParser()

    # AgentExecutor._aperform_agent_action = _aperform_agent_action
    # AgentExecutor._perform_agent_action = _perform_agent_action

    agent_executor = ZhipuAiAllToolsAgentExecutor(
        agent=agent,
        tools=tools,
        verbose=verbose,
        callbacks=callbacks,
        return_intermediate_steps=True,
    )

    return agent_executor


async def wrap_done(fn: Awaitable, event: asyncio.Event):
    """Wrap an awaitable with a event to signal when it's done or an exception is raised."""
    try:
        await fn
    except Exception as e:
        msg = f"Caught exception: {e}"
        logger.error(f"{e.__class__.__name__}: {msg}", exc_info=e)
    finally:
        # Signal the aiter to stop.
        event.set()


OutputType = Union[
    AllToolsAction,
    AllToolsActionToolStart,
    AllToolsActionToolEnd,
    AllToolsFinish,
    AllToolsLLMStatus,
]


class ZhipuAIAllToolsRunnable(RunnableSerializable[Dict, OutputType]):
    agent_executor: AgentExecutor
    """ZhipuAI AgentExecutor."""

    model_name: str = Field(default="glm-4-alltools")
    """工具模型"""
    callback: AgentExecutorAsyncIteratorCallbackHandler
    """ZhipuAI AgentExecutor callback."""
    intermediate_steps: List[Tuple[AgentAction, BaseToolOutput]] = []
    """intermediate_steps to store the data to be processed."""
    history: List[Union[List, Tuple, Dict]] = []
    """user message history"""

    class Config:
        arbitrary_types_allowed = True

    if PYDANTIC_V2:
        model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    @staticmethod
    def paser_all_tools(
        tool: Dict[str, Any], callbacks: List[BaseCallbackHandler] = []
    ) -> AdapterAllTool:
        platform_params = {}
        if tool["type"] in tool:
            platform_params = tool[tool["type"]]

        if tool["type"] in TOOL_STRUCT_TYPE_TO_TOOL_CLASS:
            all_tool = TOOL_STRUCT_TYPE_TO_TOOL_CLASS[tool["type"]](
                name=tool["type"], platform_params=platform_params, callbacks=callbacks
            )
            return all_tool
        else:
            raise ValueError(f"Unknown tool type: {tool['type']}")

    @classmethod
    def create_agent_executor(
        cls,
        model_name: str,
        *,
        intermediate_steps: List[Tuple[AgentAction, BaseToolOutput]] = [],
        history: List[Union[List, Tuple, Dict]] = [],
        tools: Sequence[
            Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool]
        ] = None,
        temperature: float = 0.7,
        **kwargs: Any,
    ) -> "ZhipuAIAllToolsRunnable":
        """Create an ZhipuAI Assistant and instantiate the Runnable."""

        callback = AgentExecutorAsyncIteratorCallbackHandler()
        callbacks = [callback]
        params = dict(
            streaming=True,
            verbose=True,
            callbacks=callbacks,
            model=model_name,
            temperature=temperature,
            **kwargs,
        )

        llm = ChatZhipuAI(**params)

        llm_with_all_tools = None

        temp_tools = []
        if tools:
            llm_with_all_tools = llm.bind(
                tools=[_get_assistants_tool(tool) for tool in tools]
            )

            temp_tools.extend(
                [
                    t.copy(update={"callbacks": callbacks})
                    for t in tools
                    if not _is_assistants_builtin_tool(t)
                ]
            )

            assistants_builtin_tools = []
            for t in tools:
                # TODO: platform tools built-in for all tools,
                #       load with langchain_glm/agents/all_tools_agent.py:108
                # AdapterAllTool implements it
                if _is_assistants_builtin_tool(t):
                    assistants_builtin_tools.append(cls.paser_all_tools(t, callbacks))
            temp_tools.extend(assistants_builtin_tools)

        agent_executor = _agents_registry(
            llm=llm,
            callbacks=callbacks,
            tools=temp_tools,
            llm_with_all_tools=llm_with_all_tools,
            verbose=True,
        )
        return cls(
            model_name=model_name,
            agent_executor=agent_executor,
            callback=callback,
            intermediate_steps=intermediate_steps,
            history=history,
            **kwargs,
        )

    def invoke(
        self, chat_input: str, config: Optional[RunnableConfig] = None
    ) -> AsyncIterable[OutputType]:
        async def chat_iterator() -> AsyncIterable[OutputType]:
            history_message = []
            if self.history:
                _history = [History.from_data(h) for h in self.history]
                chat_history = [h.to_msg_tuple() for h in _history]

                history_message = convert_to_messages(chat_history)

            task = asyncio.create_task(
                wrap_done(
                    self.agent_executor.ainvoke(
                        {
                            "input": chat_input,
                            "chat_history": history_message,
                            "agent_scratchpad": lambda x: format_to_zhipuai_all_tool_messages(
                                self.intermediate_steps
                            ),
                        }
                    ),
                    self.callback.done,
                )
            )

            async for chunk in self.callback.aiter():
                data = json.loads(chunk)
                class_status = None
                if data["status"] == AgentStatus.llm_start:
                    class_status = AllToolsLLMStatus(
                        run_id=data["run_id"],
                        status=data["status"],
                        text=data["text"],
                    )

                elif data["status"] == AgentStatus.llm_new_token:
                    class_status = AllToolsLLMStatus(
                        run_id=data["run_id"],
                        status=data["status"],
                        text=data["text"],
                    )
                elif data["status"] == AgentStatus.llm_end:
                    class_status = AllToolsLLMStatus(
                        run_id=data["run_id"],
                        status=data["status"],
                        text=data["text"],
                    )
                elif data["status"] == AgentStatus.agent_action:
                    class_status = AllToolsAction(
                        run_id=data["run_id"], status=data["status"], **data["action"]
                    )

                elif data["status"] == AgentStatus.tool_start:
                    class_status = AllToolsActionToolStart(
                        run_id=data["run_id"],
                        status=data["status"],
                        tool_input=data["tool_input"],
                        tool=data["tool"],
                    )

                elif data["status"] in [AgentStatus.tool_end]:
                    class_status = AllToolsActionToolEnd(
                        run_id=data["run_id"],
                        status=data["status"],
                        tool=data["tool"],
                        tool_output=str(data["tool_output"]),
                    )
                elif data["status"] == AgentStatus.agent_finish:
                    class_status = AllToolsFinish(
                        run_id=data["run_id"],
                        status=data["status"],
                        **data["finish"],
                    )

                elif data["status"] == AgentStatus.agent_finish:
                    class_status = AllToolsLLMStatus(
                        run_id=data["run_id"],
                        status=data["status"],
                        text=data["outputs"]["output"],
                    )

                elif data["status"] == AgentStatus.error:
                    class_status = AllToolsLLMStatus(
                        run_id=data.get("run_id", "abc"),
                        status=data["status"],
                        text=json.dumps(data, ensure_ascii=False),
                    )
                elif data["status"] == AgentStatus.chain_start:
                    class_status = AllToolsLLMStatus(
                        run_id=data["run_id"],
                        status=data["status"],
                        text="",
                    )
                elif data["status"] == AgentStatus.chain_end:
                    class_status = AllToolsLLMStatus(
                        run_id=data["run_id"],
                        status=data["status"],
                        text=data["outputs"]["output"],
                    )

                yield class_status

            await task

            if self.callback.out:
                self.history.append({"role": "user", "content": chat_input})
                self.history.append(
                    {"role": "assistant", "content": self.callback.outputs["output"]}
                )
                self.intermediate_steps.extend(self.callback.intermediate_steps)

        return chat_iterator()



# -*- coding: utf-8 -*-
import json
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from langchain_core.agents import AgentAction
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)

from langchain_glm.agent_toolkits import AdapterAllTool
from langchain_glm.agent_toolkits.all_tools.tool import (
    AllToolExecutor,
    BaseToolOutput,
)

logger = logging.getLogger(__name__)


class CodeInterpreterToolOutput(BaseToolOutput):
    platform_params: Dict[str, Any]
    tool: str
    code_input: str
    code_output: Dict[str, Any]

    def __init__(
        self,
        tool: str,
        code_input: str,
        code_output: Dict[str, Any],
        platform_params: Dict[str, Any],
        **extras: Any,
    ) -> None:
        data = CodeInterpreterToolOutput.paser_data(
            tool=tool, code_input=code_input, code_output=code_output
        )
        super().__init__(data, "", "", **extras)
        self.platform_params = platform_params
        self.tool = tool
        self.code_input = code_input
        self.code_output = code_output

    @staticmethod
    def paser_data(tool: str, code_input: str, code_output: Dict[str, Any]) -> str:
        return f"""Access：{tool}, Message: {code_input},{code_output}"""


@dataclass
class CodeInterpreterAllToolExecutor(AllToolExecutor):
    """platform adapter tool for code interpreter tool"""

    name: str

    @staticmethod
    def _python_ast_interpreter(
        code_input: str, platform_params: Dict[str, Any] = None
    ):
        """Use Shell to execute system shell commands"""

        try:
            from langchain_experimental.tools import PythonAstREPLTool

            tool = PythonAstREPLTool()
            out = tool.run(tool_input=code_input)
            if str(out) == "":
                raise ValueError(f"Tool {tool.name} local sandbox is out empty")
            return CodeInterpreterToolOutput(
                tool=tool.name,
                code_input=code_input,
                code_output=out,
                platform_params=platform_params,
            )
        except ImportError:
            raise AttributeError(
                "This tool has been moved to langchain experiment. "
                "This tool has access to a python REPL. "
                "For best practices make sure to sandbox this tool. "
                "Read https://github.com/langchain-ai/langchain/blob/master/SECURITY.md "
                "To keep using this code as is, install langchain experimental and "
                "update relevant imports replacing 'langchain' with 'langchain_experimental'"
            )

    def run(
        self,
        tool: str,
        tool_input: str,
        log: str,
        outputs: List[Union[str, dict]] = None,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> CodeInterpreterToolOutput:
        if outputs is None or str(outputs).strip() == "":
            if "auto" == self.platform_params.get("sandbox", "auto"):
                raise ValueError(
                    f"Tool {self.name} sandbox is auto , but log is None, is server error"
                )
            elif "none" == self.platform_params.get("sandbox", "auto"):
                logger.warning(
                    f"Tool {self.name} sandbox is local!!!, this not safe, please use jupyter sandbox it"
                )
                return self._python_ast_interpreter(
                    code_input=tool_input, platform_params=self.platform_params
                )

        return CodeInterpreterToolOutput(
            tool=tool,
            code_input=tool_input,
            code_output=json.dumps(outputs),
            platform_params=self.platform_params,
        )

    async def arun(
        self,
        tool: str,
        tool_input: str,
        log: str,
        outputs: List[Union[str, dict]] = None,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> CodeInterpreterToolOutput:
        """Use the tool asynchronously."""
        if outputs is None or str(outputs).strip() == "" or len(outputs) == 0:
            if "auto" == self.platform_params.get("sandbox", "auto"):
                raise ValueError(
                    f"Tool {self.name} sandbox is auto , but log is None, is server error"
                )
            elif "none" == self.platform_params.get("sandbox", "auto"):
                logger.warning(
                    f"Tool {self.name} sandbox is local!!!, this not safe, please use jupyter sandbox it"
                )
                return self._python_ast_interpreter(
                    code_input=tool_input, platform_params=self.platform_params
                )

        return CodeInterpreterToolOutput(
            tool=tool,
            code_input=tool_input,
            code_output=json.dumps(outputs),
            platform_params=self.platform_params,
        )


class CodeInterpreterAdapterAllTool(AdapterAllTool[CodeInterpreterAllToolExecutor]):
    @classmethod
    def get_type(cls) -> str:
        return "CodeInterpreterAdapterAllTool"

    def _build_adapter_all_tool(
        self, platform_params: Dict[str, Any]
    ) -> CodeInterpreterAllToolExecutor:
        return CodeInterpreterAllToolExecutor(
            name="code_interpreter", platform_params=platform_params
        )



# -*- coding: utf-8 -*-
from langchain_glm.agents.zhipuai_all_tools.base import (
    ZhipuAIAllToolsRunnable,
)
from langchain_glm.agents.zhipuai_all_tools.schema import (
    AllToolsAction,
    AllToolsActionToolEnd,
    AllToolsActionToolStart,
    AllToolsBaseComponent,
    AllToolsFinish,
    AllToolsLLMStatus,
    MsgType,
)

__all__ = [
    "ZhipuAIAllToolsRunnable",
    "MsgType",
    "AllToolsBaseComponent",
    "AllToolsAction",
    "AllToolsFinish",
    "AllToolsActionToolStart",
    "AllToolsActionToolEnd",
    "AllToolsLLMStatus",
]



# -*- coding: utf-8 -*-
"""platform adapter tool """

from __future__ import annotations

import json
import logging
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from dataclasses_json import DataClassJsonMixin
from langchain_core.agents import AgentAction, AgentFinish, AgentStep
from langchain_core.callbacks import (
    AsyncCallbackManagerForChainRun,
    AsyncCallbackManagerForToolRun,
    BaseCallbackManager,
    CallbackManagerForChainRun,
    CallbackManagerForToolRun,
    Callbacks,
)
from langchain_core.tools import BaseTool

from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agents.output_parsers.code_interpreter import (
    CodeInterpreterAgentAction,
)
from langchain_glm.agents.output_parsers.drawing_tool import DrawingToolAgentAction
from langchain_glm.agents.output_parsers.web_browser import WebBrowserAgentAction

logger = logging.getLogger(__name__)


class BaseToolOutput:
    """
    LLM 要求 Tool 的输出为 str，但 Tool 用在别处时希望它正常返回结构化数据。
    只需要将 Tool 返回值用该类封装，能同时满足两者的需要。
    基类简单的将返回值字符串化，或指定 format="json" 将其转为 json。
    用户也可以继承该类定义自己的转换方法。
    """

    def __init__(
        self,
        data: Any,
        format: str = "",
        data_alias: str = "",
        **extras: Any,
    ) -> None:
        self.data = data
        self.format = format
        self.extras = extras
        if data_alias:
            setattr(self, data_alias, property(lambda obj: obj.data))

    def __str__(self) -> str:
        if self.format == "json":
            return json.dumps(self.data, ensure_ascii=False, indent=2)
        else:
            return str(self.data)


@dataclass
class AllToolExecutor(DataClassJsonMixin):
    platform_params: Dict[str, Any]

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> BaseToolOutput:
        pass

    @abstractmethod
    async def arun(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> BaseToolOutput:
        pass


E = TypeVar("E", bound=AllToolExecutor)


class AdapterAllTool(BaseTool, Generic[E]):
    """platform adapter tool for all tools."""

    name: str
    description: str

    platform_params: Dict[str, Any]
    """tools params """
    adapter_all_tool: E

    def __init__(self, name: str, platform_params: Dict[str, Any], **data: Any):
        super().__init__(
            name=name,
            description=f"platform adapter tool for {name}",
            platform_params=platform_params,
            adapter_all_tool=self._build_adapter_all_tool(platform_params),
            **data,
        )

    @abstractmethod
    def _build_adapter_all_tool(self, platform_params: Dict[str, Any]) -> E:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_type(cls) -> str:
        raise NotImplementedError

    def _to_args_and_kwargs(self, tool_input: Union[str, Dict]) -> Tuple[Tuple, Dict]:
        # For backwards compatibility, if run_input is a string,
        # pass as a positional argument.
        if tool_input is None:
            return (), {}
        if isinstance(tool_input, str):
            return (tool_input,), {}
        else:
            # for tool defined with `*args` parameters
            # the args_schema has a field named `args`
            # it should be expanded to actual *args
            # e.g.: test_tools
            #       .test_named_tool_decorator_return_direct
            #       .search_api
            if "args" in tool_input:
                args = tool_input["args"]
                if args is None:
                    tool_input.pop("args")
                    return (), tool_input
                elif isinstance(args, tuple):
                    tool_input.pop("args")
                    return args, tool_input
            return (), tool_input

    def _run(
        self,
        agent_action: AgentAction,
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
        **tool_run_kwargs: Any,
    ) -> Any:
        if (
            AdapterAllToolStructType.CODE_INTERPRETER == agent_action.tool
            and isinstance(agent_action, CodeInterpreterAgentAction)
        ):
            return self.adapter_all_tool.run(
                **{
                    "tool": agent_action.tool,
                    "tool_input": agent_action.tool_input,
                    "log": agent_action.log,
                    "outputs": agent_action.outputs,
                },
                **tool_run_kwargs,
            )
        elif AdapterAllToolStructType.DRAWING_TOOL == agent_action.tool and isinstance(
            agent_action, DrawingToolAgentAction
        ):
            return self.adapter_all_tool.run(
                **{
                    "tool": agent_action.tool,
                    "tool_input": agent_action.tool_input,
                    "log": agent_action.log,
                    "outputs": agent_action.outputs,
                },
                **tool_run_kwargs,
            )
        elif AdapterAllToolStructType.WEB_BROWSER == agent_action.tool and isinstance(
            agent_action, WebBrowserAgentAction
        ):
            return self.adapter_all_tool.run(
                **{
                    "tool": agent_action.tool,
                    "tool_input": agent_action.tool_input,
                    "log": agent_action.log,
                    "outputs": agent_action.outputs,
                },
                **tool_run_kwargs,
            )
        else:
            raise KeyError()

    async def _arun(
        self,
        agent_action: AgentAction,
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
        **tool_run_kwargs: Any,
    ) -> Any:
        if (
            AdapterAllToolStructType.CODE_INTERPRETER == agent_action.tool
            and isinstance(agent_action, CodeInterpreterAgentAction)
        ):
            return await self.adapter_all_tool.arun(
                **{
                    "tool": agent_action.tool,
                    "tool_input": agent_action.tool_input,
                    "log": agent_action.log,
                    "outputs": agent_action.outputs,
                },
                **tool_run_kwargs,
            )

        elif AdapterAllToolStructType.DRAWING_TOOL == agent_action.tool and isinstance(
            agent_action, DrawingToolAgentAction
        ):
            return await self.adapter_all_tool.arun(
                **{
                    "tool": agent_action.tool,
                    "tool_input": agent_action.tool_input,
                    "log": agent_action.log,
                    "outputs": agent_action.outputs,
                },
                **tool_run_kwargs,
            )
        elif AdapterAllToolStructType.WEB_BROWSER == agent_action.tool and isinstance(
            agent_action, WebBrowserAgentAction
        ):
            return await self.adapter_all_tool.arun(
                **{
                    "tool": agent_action.tool,
                    "tool_input": agent_action.tool_input,
                    "log": agent_action.log,
                    "outputs": agent_action.outputs,
                },
                **tool_run_kwargs,
            )
        else:
            raise KeyError()



# -*- coding: utf-8 -*-
import json
import uuid
from abc import abstractmethod
from enum import Enum, auto
from numbers import Number
from typing import Any, Dict, List, Optional, Union

from typing_extensions import ClassVar, Self
from zhipuai.core import PYDANTIC_V2, BaseModel, ConfigDict


class MsgType:
    TEXT = 1
    IMAGE = 2
    AUDIO = 3
    VIDEO = 4


class AllToolsBaseComponent(BaseModel):
    if PYDANTIC_V2:
        model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)
    else:

        class Config:
            arbitrary_types_allowed = True

    @classmethod
    @abstractmethod
    def class_name(cls) -> str:
        """Get class name."""

    def to_dict(self, **kwargs: Any) -> Dict[str, Any]:
        data = self.dict(**kwargs)
        data["class_name"] = self.class_name()
        return data

    def to_json(self, **kwargs: Any) -> str:
        data = self.to_dict(**kwargs)
        return json.dumps(data, ensure_ascii=False)

    # TODO: return type here not supported by current mypy version
    @classmethod
    def from_dict(cls, data: Dict[str, Any], **kwargs: Any) -> Self:  # type: ignore
        if isinstance(kwargs, dict):
            data.update(kwargs)

        data.pop("class_name", None)
        return cls(**data)

    @classmethod
    def from_json(cls, data_str: str, **kwargs: Any) -> Self:  # type: ignore
        data = json.loads(data_str)
        return cls.from_dict(data, **kwargs)


class AllToolsAction(AllToolsBaseComponent):
    """AgentFinish with run and thread metadata."""

    run_id: str
    status: int  # AgentStatus
    tool: str
    tool_input: Union[str, Dict[str, str], Dict[str, Number]]
    log: str

    @classmethod
    def class_name(cls) -> str:
        return "AllToolsAction"


class AllToolsFinish(AllToolsBaseComponent):
    """AgentFinish with run and thread metadata."""

    run_id: str
    status: int  # AgentStatus
    return_values: Dict[str, str]
    log: str

    @classmethod
    def class_name(cls) -> str:
        return "AllToolsFinish"


class AllToolsActionToolStart(AllToolsBaseComponent):
    """AllToolsAction with run and thread metadata."""

    run_id: str
    status: int  # AgentStatus
    tool: str
    tool_input: Optional[str] = None

    @classmethod
    def class_name(cls) -> str:
        return "AllToolsActionToolStart"


class AllToolsActionToolEnd(AllToolsBaseComponent):
    """AllToolsActionToolEnd with run and thread metadata."""

    run_id: str

    status: int  # AgentStatus
    tool: str
    tool_output: str

    @classmethod
    def class_name(cls) -> str:
        return "AllToolsActionToolEnd"


class AllToolsLLMStatus(AllToolsBaseComponent):
    run_id: str
    status: int  # AgentStatus
    text: str
    message_type: int = MsgType.TEXT

    @classmethod
    def class_name(cls) -> str:
        return "AllToolsLLMStatus"



# -*- coding: utf-8 -*-
import json
from typing import List, Sequence, Tuple, Union

from langchain.agents.output_parsers.tools import ToolAgentAction
from langchain_core.agents import AgentAction
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ToolMessage,
)

from langchain_glm.agent_toolkits import BaseToolOutput
from langchain_glm.agent_toolkits.all_tools.code_interpreter_tool import (
    CodeInterpreterToolOutput,
)
from langchain_glm.agent_toolkits.all_tools.drawing_tool import DrawingToolOutput
from langchain_glm.agent_toolkits.all_tools.web_browser_tool import (
    WebBrowserToolOutput,
)
from langchain_glm.agents.output_parsers.code_interpreter import (
    CodeInterpreterAgentAction,
)
from langchain_glm.agents.output_parsers.drawing_tool import DrawingToolAgentAction
from langchain_glm.agents.output_parsers.web_browser import WebBrowserAgentAction


def _create_tool_message(
    agent_action: ToolAgentAction, observation: Union[str, BaseToolOutput]
) -> ToolMessage:
    """Convert agent action and observation into a function message.
    Args:
        agent_action: the tool invocation request from the agent
        observation: the result of the tool invocation
    Returns:
        FunctionMessage that corresponds to the original tool invocation
    """
    if not isinstance(observation, str):
        try:
            content = json.dumps(observation, ensure_ascii=False)
        except Exception:
            content = str(observation)
    else:
        content = observation
    return ToolMessage(
        tool_call_id=agent_action.tool_call_id,
        content=content,
        additional_kwargs={"name": agent_action.tool},
    )


def format_to_zhipuai_all_tool_messages(
    intermediate_steps: Sequence[Tuple[AgentAction, BaseToolOutput]],
) -> List[BaseMessage]:
    """Convert (AgentAction, tool output) tuples into FunctionMessages.

    Args:
        intermediate_steps: Steps the LLM has taken to date, along with observations

    Returns:
        list of messages to send to the LLM for the next prediction

    """
    messages = []
    for agent_action, observation in intermediate_steps:
        if isinstance(agent_action, CodeInterpreterAgentAction):
            if isinstance(observation, CodeInterpreterToolOutput):
                if "auto" == observation.platform_params.get("sandbox", "auto"):
                    new_messages = [
                        AIMessage(content=str(observation.code_input)),
                        _create_tool_message(agent_action, observation),
                    ]

                    messages.extend(
                        [new for new in new_messages if new not in messages]
                    )
                elif "none" == observation.platform_params.get("sandbox", "auto"):
                    new_messages = [
                        AIMessage(content=str(observation.code_input)),
                        _create_tool_message(agent_action, observation.code_output),
                    ]

                    messages.extend(
                        [new for new in new_messages if new not in messages]
                    )
                else:
                    raise ValueError(
                        f"Unknown sandbox type: {observation.platform_params.get('sandbox', 'auto')}"
                    )
            else:
                raise ValueError(f"Unknown observation type: {type(observation)}")

        elif isinstance(agent_action, DrawingToolAgentAction):
            if isinstance(observation, DrawingToolOutput):
                new_messages = [AIMessage(content=str(observation))]
                messages.extend([new for new in new_messages if new not in messages])
            else:
                raise ValueError(f"Unknown observation type: {type(observation)}")

        elif isinstance(agent_action, WebBrowserAgentAction):
            if isinstance(observation, WebBrowserToolOutput):
                new_messages = [AIMessage(content=str(observation))]
                messages.extend([new for new in new_messages if new not in messages])
            else:
                raise ValueError(f"Unknown observation type: {type(observation)}")

        elif isinstance(agent_action, ToolAgentAction):
            ai_msgs = AIMessage(
                content=f"arguments='{agent_action.tool_input}', name='{agent_action.tool}'"
            )
            new_messages = [ai_msgs, _create_tool_message(agent_action, observation)]
            messages.extend([new for new in new_messages if new not in messages])
        else:
            messages.append(AIMessage(content=agent_action.log))
    return messages



# -*- coding: utf-8 -*-
"""**Callback handlers** allow listening to events in LangChain.

**Class hierarchy:**

.. code-block::

    BaseCallbackHandler --> <name>CallbackHandler  # Example: AimCallbackHandler
"""
from langchain_glm.callbacks.agent_callback_handler import (
    AgentExecutorAsyncIteratorCallbackHandler,
)

__all__ = [
    "AgentExecutorAsyncIteratorCallbackHandler",
]



# -*- coding: utf-8 -*-
from typing import Dict, Type

from langchain_glm.agent_toolkits import AdapterAllTool
from langchain_glm.agent_toolkits.all_tools.code_interpreter_tool import (
    CodeInterpreterAdapterAllTool,
)
from langchain_glm.agent_toolkits.all_tools.drawing_tool import (
    DrawingAdapterAllTool,
)
from langchain_glm.agent_toolkits.all_tools.struct_type import (
    AdapterAllToolStructType,
)
from langchain_glm.agent_toolkits.all_tools.web_browser_tool import (
    WebBrowserAdapterAllTool,
)

TOOL_STRUCT_TYPE_TO_TOOL_CLASS: Dict[AdapterAllToolStructType, Type[AdapterAllTool]] = {
    AdapterAllToolStructType.CODE_INTERPRETER: CodeInterpreterAdapterAllTool,
    AdapterAllToolStructType.DRAWING_TOOL: DrawingAdapterAllTool,
    AdapterAllToolStructType.WEB_BROWSER: WebBrowserAdapterAllTool,
}
